import json
import logging

logger = logging.getLogger(__name__)


class SpytEnablers(object):
    BYOP_KEY = "spark.hadoop.yt.byop.enabled"
    PROFILING_KEY = "spark.hadoop.yt.profiling.enabled"
    ARROW_KEY = "spark.hadoop.yt.read.arrow.enabled"
    MTN_KEY = "spark.hadoop.yt.mtn.enabled"
    YT_METRICS_KEY = "spark.ytsaurus.metrics.enabled"
    IPV6_KEY = "spark.hadoop.yt.preferenceIpv6.enabled"
    TCP_PROXY_KEY = "spark.hadoop.yt.tcpProxy.enabled"
    SQUASHFS_KEY = "spark.ytsaurus.squashfs.enabled"

    def __init__(self, enable_byop=False, enable_profiling=False, enable_arrow=True,
                 enable_mtn=False, enable_yt_metrics=True, enable_preference_ipv6=True,
                 enable_tcp_proxy=False, enable_squashfs=False):
        self.enable_byop = enable_byop
        self.enable_profiling = enable_profiling
        self.enable_arrow = enable_byop if enable_arrow is None else enable_arrow
        self.enable_mtn = enable_mtn
        self.enable_yt_metrics = enable_yt_metrics
        self.enable_preference_ipv6 = enable_preference_ipv6
        self.enable_tcp_proxy = enable_tcp_proxy
        self.enable_squashfs = enable_squashfs
        self.config_enablers = {}

    def _get_enabler(self, enabler, old_enabler_name, enabler_name):
        config_enabler = False
        # COMPAT(alex-shishkin): Remove when there is no cluster under 1.76.0
        if old_enabler_name in self.config_enablers:
            config_enabler = self.config_enablers.get(old_enabler_name)
        elif enabler_name in self.config_enablers:
            config_enabler = self.config_enablers.get(enabler_name)
        if enabler and not config_enabler:
            logger.warning("{} is enabled in start arguments, but disabled for current cluster version. "
                           "It will be disabled".format(old_enabler_name))
        return enabler and config_enabler

    def apply_config(self, config):
        self.config_enablers = config.get("enablers") or {}
        self.enable_byop = self._get_enabler(self.enable_byop, "enable_byop", self.BYOP_KEY)
        self.enable_mtn = self._get_enabler(self.enable_mtn, "enable_mtn", self.MTN_KEY)
        self.enable_yt_metrics = self._get_enabler(self.enable_yt_metrics, "enable_solomon_agent",
                                                   self.YT_METRICS_KEY)
        self.enable_tcp_proxy = self._get_enabler(self.enable_tcp_proxy, "enable_tcp_proxy", self.TCP_PROXY_KEY)
        self.enable_squashfs = self._get_enabler(self.enable_squashfs, "enable_squashfs", self.SQUASHFS_KEY)

    def __str__(self):
        return json.dumps(self.get_conf())

    def get_conf(self):
        return {
            self.BYOP_KEY: str(self.enable_byop),
            self.ARROW_KEY: str(self.enable_arrow),
            self.IPV6_KEY: str(self.enable_preference_ipv6),
        }



import logging
import os
import sys
import shutil
import signal
import tempfile
import time
from contextlib import contextmanager
from py4j.java_gateway import JavaGateway, GatewayParameters
from pyspark.serializers import read_int, UTF8Deserializer
from subprocess import Popen, PIPE
from py4j.protocol import Py4JJavaError
from enum import Enum
from datetime import timedelta
from .utils import scala_buffer_to_list, get_spark_home, get_spyt_home, get_spyt_conf_dir


logger = logging.getLogger(__name__)


def launch_gateway(memory="512m",
                   java_home=None,
                   java_opts=None,
                   additional_jars=None,
                   additional_environ=None,
                   prefer_ipv6=False):  # Internal Yandex users must enable ipv6 option by default
    spark_home = get_spark_home()
    spyt_home = get_spyt_home()
    java = os.path.join(java_home, "bin", "java") if java_home else "java"
    additional_jars = additional_jars or []

    command = [java, "-Xmx{}".format(memory)]
    command += java_opts or []
    if prefer_ipv6:
        command.append('-Djava.net.preferIPv6Addresses=true')
    spark_patch = [
        os.path.join(spyt_home, 'jars', jar)
        for jar in os.listdir(os.path.join(spyt_home, 'jars'))
        if 'spark-yt-spark-patch' in jar][0]
    command += [
        f"-javaagent:{spark_patch}",
        "-XX:+IgnoreUnrecognizedVMOptions",
        "--add-opens=java.base/java.lang=ALL-UNNAMED",
        "-cp", ":".join(additional_jars + _submit_classpath(spark_home)),
        "tech.ytsaurus.spyt.submit.PythonGatewayServer"
    ]

    conn_info_dir = tempfile.mkdtemp()
    try:
        fd, conn_info_file = tempfile.mkstemp(dir=conn_info_dir)
        os.close(fd)
        os.unlink(conn_info_file)

        env = dict(os.environ)
        env.update(additional_environ or {})
        env["_SPYT_SUBMIT_CONN_INFO_PATH"] = conn_info_file
        env["SPARK_HOME"] = spark_home
        env["SPARK_CONF_DIR"] = get_spyt_conf_dir()

        # Launch the Java gateway.
        popen_kwargs = {'stdin': PIPE, 'env': env}

        # Don't send ctrl-c / SIGINT to the Java gateway:
        def preexec_func():
            signal.signal(signal.SIGINT, signal.SIG_IGN)

        popen_kwargs['preexec_fn'] = preexec_func
        logger.debug(f"Starting JVM process. Path to bin: {java}")
        proc = Popen(command, **popen_kwargs)

        # Wait for the file to appear, or for the process to exit, whichever happens first.
        while not proc.poll() and not os.path.isfile(conn_info_file):
            time.sleep(0.1)

        if not os.path.isfile(conn_info_file):
            raise Exception("Java gateway process exited before sending its port number")

        logger.debug("Process started. Reading gateway data")
        with open(conn_info_file, "rb") as info:
            gateway_port = read_int(info)
            gateway_secret = UTF8Deserializer().loads(info)
    finally:
        shutil.rmtree(conn_info_dir)

    # Connect to the gateway (or client server to pin the thread between JVM and Python)
    address = '::1' if prefer_ipv6 else '127.0.0.1'
    logger.debug(f"Connecting to created gateway on {address}:{gateway_port}")
    gateway = JavaGateway(
        gateway_parameters=GatewayParameters(address=address, port=gateway_port,
                                             auth_token=gateway_secret,
                                             auto_convert=True))

    logger.debug("Gateway connection established")
    # Store a reference to the Popen object for use by the caller (e.g., in reading stdout/stderr)
    gateway.proc = proc
    return gateway


def _submit_classpath(spark_home=None):
    spark_home = spark_home or get_spark_home()
    spyt_home = get_spyt_home()
    jars_cp = [os.path.join(home, "jars/*") for home in [spyt_home, spark_home]]
    return [os.path.join(spyt_home, "conf")] + jars_cp


def shutdown_gateway(gateway):
    gateway.shutdown()
    gateway.proc.stdin.close()
    logger.debug("Gateway stopped")


@contextmanager
def java_gateway(kill_jvm_on_shutdown=True, *args, **kwargs):
    logger.debug("Launching java gateway")
    gateway = launch_gateway(*args, **kwargs)
    try:
        yield gateway
    except Py4JJavaError as e:
        raise RuntimeError(str(e))
    finally:
        shutdown_gateway(gateway)
        if kill_jvm_on_shutdown:
            gateway.proc.kill()
            logger.debug("Java process stopped")


class RetryConfig(object):
    def __init__(self, enable_retry=True, retry_limit=10, retry_interval=timedelta(minutes=1)):
        self.enable_retry = enable_retry
        self.retry_limit = retry_limit
        self.retry_interval = retry_interval
        self.wait_submission_id_retry_limit = None  # Every retry will be 5 seconds

    def _to_java(self, gateway):
        jduration = gateway.jvm.tech.ytsaurus.spyt.submit.RetryConfig.durationFromSeconds(
            int(self.retry_interval.total_seconds()))
        if self.wait_submission_id_retry_limit:  # COMPAT(alex-shishkin)
            args = [self.enable_retry, self.retry_limit, jduration, self.wait_submission_id_retry_limit]
        else:
            args = [self.enable_retry, self.retry_limit, jduration]
        return gateway.jvm.tech.ytsaurus.spyt.submit.RetryConfig(*args)


class SparkSubmissionClient(object):
    def __new__(cls, *args, **kwargs):
        instance = super(SparkSubmissionClient, cls).__new__(cls)
        return instance

    def __init__(self, gateway, proxy, discovery_path, user, token):
        self._jclient = gateway.jvm.tech.ytsaurus.spyt.submit.SubmissionClient(proxy, discovery_path, user, token)
        self.gateway = gateway

    def new_launcher(self):
        jlauncher = self._jclient.newLauncher()
        return SparkLauncher(jlauncher, self.gateway)

    def submit(self, launcher, retry_config=None):
        if retry_config is None:
            retry_config = RetryConfig()
        logger.debug("Job submit")
        jresult = self._jclient.submit(launcher._jlauncher, retry_config._to_java(self.gateway))
        if jresult.isFailure():
            raise RuntimeError(jresult.failed().get().getMessage())
        return jresult.get()

    def kill(self, submission_id):
        return self._jclient.kill(submission_id)

    def get_active_drivers(self):
        return scala_buffer_to_list(self._jclient.getActiveDrivers())

    def get_completed_drivers(self):
        return scala_buffer_to_list(self._jclient.getCompletedDrivers())

    def get_all_drivers(self):
        return scala_buffer_to_list(self._jclient.getAllDrivers())

    def get_status(self, submission_id):
        # TODO refactor to use requests package
        return SubmissionStatus.from_string(self._jclient.getStringStatus(submission_id))

    def get_app_status(self, driver_id):
        # TODO refactor to use requests package
        return ApplicationStatus.from_string(self._jclient.getStringApplicationStatus(driver_id))

    def wait_final(self, app_id, ping_period=5):
        logger.debug(f"Waiting app {app_id} finishing")
        status = self.get_status(app_id)
        while not SubmissionStatus.is_final(status):
            status = self.get_status(app_id)
            logger.debug(f"Current app {app_id} status: {status}")
            time.sleep(ping_period)
        return status


class SubmissionStatus(Enum):
    SUBMITTED = "SUBMITTED"
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    RELAUNCHING = "RELAUNCHING"
    UNKNOWN = "UNKNOWN"
    KILLED = "KILLED"
    FAILED = "FAILED"
    ERROR = "ERROR"
    UNDEFINED = "UNDEFINED"

    @staticmethod
    def from_string(name):
        return next(member for n, member in SubmissionStatus.__members__.items() if n == name)

    @staticmethod
    def is_final(status):
        return (
            status is SubmissionStatus.FINISHED or
            status is SubmissionStatus.UNKNOWN or
            status is SubmissionStatus.KILLED or
            status is SubmissionStatus.FAILED or
            status is SubmissionStatus.ERROR
        )

    @staticmethod
    def is_success(status):
        return (
            status is SubmissionStatus.FINISHED
        )

    @staticmethod
    def is_failure(status):
        return (
            status is SubmissionStatus.UNKNOWN or
            status is SubmissionStatus.KILLED or
            status is SubmissionStatus.FAILED or
            status is SubmissionStatus.ERROR
        )


class ApplicationStatus(Enum):
    WAITING = "WAITING"
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    FAILED = "FAILED"
    KILLED = "KILLED"
    UNKNOWN = "UNKNOWN"
    UNDEFINED = "UNDEFINED"

    @staticmethod
    def from_string(name):
        return next(member for n, member in ApplicationStatus.__members__.items() if n == name)

    @staticmethod
    def is_final(status):
        return (
            status is ApplicationStatus.FINISHED or
            status is ApplicationStatus.UNKNOWN or
            status is ApplicationStatus.KILLED or
            status is ApplicationStatus.FAILED
        )

    @staticmethod
    def is_success(status):
        return (
            status is ApplicationStatus.FINISHED
        )

    @staticmethod
    def is_failure(status):
        return (
            status is ApplicationStatus.UNKNOWN or
            status is ApplicationStatus.KILLED or
            status is ApplicationStatus.FAILED
        )

    @staticmethod
    def is_waiting(status):
        return (
            status is ApplicationStatus.WAITING
        )


# wraps methods from
# https://github.com/apache/spark/blob/master/launcher/src/main/java/org/apache/spark/launcher/AbstractLauncher.java
class SparkLauncher(object):
    def __new__(cls, *args, **kwargs):
        instance = super(SparkLauncher, cls).__new__(cls)
        return instance

    def __init__(self, jlauncher, gateway):
        self._jlauncher = jlauncher
        self._jutils = gateway.jvm.tech.ytsaurus.spyt.submit.InProcessLauncherPythonUtils

    def set_app_resource(self, resource: str):
        self._jlauncher.setAppResource(resource)
        return self

    def set_main_class(self, main_class: str):
        self._jlauncher.setMainClass(main_class)
        return self

    def set_conf(self, key: str, value: str):
        self._jlauncher.setConf(key, value)
        return self

    def set_properties_file(self, path: str):
        self._jlauncher.setPropertiesFile(path)
        return self

    def set_app_name(self, app_name: str):
        self._jlauncher.setAppName(app_name)
        return self

    def add_spark_arg(self, arg: str, value: str = None):
        if value:
            self._jlauncher.addSparkArg(arg, value)
        else:
            self._jlauncher.addSparkArg(arg)
        return self

    def add_app_args(self, *args):
        for arg in args:
            self._jutils.addAppArg(self._jlauncher, arg)
        return self

    def add_jar(self, jar: str):
        self._jlauncher.addJar(jar)
        return self

    def add_file(self, file: str):
        self._jlauncher.addFile(file)
        return self

    def add_py_file(self, file: str):
        self._jlauncher.addPyFile(file)
        return self

    def set_verbose(self, verbose: bool):
        self._jlauncher.setVerbose(verbose)
        return self

    def add_spark_args(self, *spark_args):
        for i in range(0, len(spark_args), 2):
            self.add_spark_arg(spark_args[i], spark_args[i+1])
        return self

    def set_spark_conf(self, conf: dict):
        for key, value in conf.items():
            self.set_conf(key, str(value))
        return self

    def set_master(self, master: str):
        self._jlauncher.setMaster(master)
        return self

    def set_deploy_mode(self, deploy_mode: str):
        self._jlauncher.setDeployMode(deploy_mode)
        return self

    def launcher(self):
        return self._jlauncher


def create_base_spark_env(spark_home):
    spark_env = os.environ.copy()
    spark_env["SPARK_CONF_DIR"] = get_spyt_conf_dir()
    if spark_home:
        spark_env["SPARK_HOME"] = spark_home
    return spark_env


def direct_submit(yt_proxy, num_executors, main_file, deploy_mode="cluster", pool=None,
                  spark_base_args=[], job_args=[], spark_conf={}, java_home=None, prefer_ipv6=False,
                  timeout_sec=30, max_attempts=1, base_delay=10, max_delay=300):
    """
    Submits a Spark job directly to YTsaurus using the provided parameters.

    :param yt_proxy: YTsaurus proxy address (e.g., "hume.yt.yandex.net").
    :param num_executors: Number of Spark executors to use for the job.
    :param main_file: path to the main Spark application file (e.g., Python script).
    :param deploy_mode: deployment mode for Spark ("client" or "cluster", default: "cluster",
    "client" is not recommended, use direct_spark_session instead)
    :param pool: YTsaurus pool to execute this job (default: None)
    :param spark_base_args: additional Spark arguments
    :param job_args: job arguments
    :param spark_conf: additional Spark configuration as Python dict
    :param java_home: path to the Java home directory (default: None)
    :param prefer_ipv6:prefer IPv6 addresses (internal Yandex users must enable ipv6 option by default)
    :param timeout_sec: timeout for submitting the job in seconds (default: 30 sec)
    :param max_attempts: maximum number of attempts to submit the job (default: 1)
    :param base_delay: base delay between attempts in seconds, doubles for every new attempt (default: 10 sec)
    :param max_delay: maximum delay between attempts in seconds (default: 300 sec)
    :return: operation ID of the submitted Spark job.
    """
    if max_attempts < 1:
        raise ValueError("max_attempts must be greater than 0")
    if base_delay < 0:
        raise ValueError("base_delay must be non-negative")
    if max_delay < base_delay:
        raise ValueError("max_delay must be greater than or equal to base_delay")

    spark_args = []
    spark_args.extend(["--num-executors", str(num_executors)])
    if pool:
        spark_args.extend(["--queue", pool])
    spark_args.extend(spark_base_args)

    with (java_gateway(java_home=java_home, prefer_ipv6=prefer_ipv6) as gateway):
        j_launcher = gateway.jvm.org.apache.spark.launcher.InProcessLauncher()
        spark_launcher = (SparkLauncher(j_launcher, gateway)
                          .set_master("ytsaurus://" + yt_proxy)
                          .set_deploy_mode(deploy_mode)
                          .set_app_resource(main_file)
                          .set_spark_conf(spark_conf)
                          .add_app_args(*job_args)
                          .add_spark_args(*spark_args))

        j_directsubmitter = gateway.jvm.tech.ytsaurus.spyt.submit.DirectSubmitter()
        delay = base_delay
        for attempt in range(max_attempts):
            try_operation_id = j_directsubmitter.submit(spark_launcher.launcher(), timeout_sec)
            if try_operation_id.isSuccess():
                return try_operation_id.get()
            failure = try_operation_id.failed().get()
            last_error = f"Failed to submit Spark job: {failure.toString()}, {failure.getMessage()}"
            if attempt < max_attempts - 1:
                logging.warning(f"Attempt {attempt + 1}. {last_error}. Retrying in {delay} sec ...")
                time.sleep(delay)
                delay = min(delay * 2, max_delay)
        raise Exception(f"All {max_attempts} attempts failed. Last error: {last_error}")


def direct_submit_binary(yt_proxy, num_executors, spyt_version, driver_entry_point=None, executable=sys.executable,
                         deploy_mode="cluster", pool=None, spark_base_args=[], job_args=[], spark_conf={}):
    """
    Submits Spark job compiled as single binary executable file. By default, it submits itself assuming
    that this method is called from inside such binary

    :param yt_proxy: YTsaurus proxy address
    :param num_executors: number of Spark executors
    :param spyt_version: SPYT version to be used on cluster
    :param driver_entry_point: Spark driver entry point (default: None)
    :param executable: (default: sys.executable)
    :param deploy_mode: Spark deploy mode, client or cluster (default: cluster)
    :param pool: YTsaurus pool to execute this job (default: None)
    :param spark_base_args: additional Spark arguments
    :param job_args: job arguments
    :param spark_conf: additional Spark configuration as Python dict
    :return:
    """
    conf = dict(spark_conf)
    conf["spark.ytsaurus.spyt.version"] = spyt_version
    if driver_entry_point:
        conf["spark.ytsaurus.python.binary.entry.point"] = driver_entry_point

    return direct_submit(yt_proxy, num_executors, executable, deploy_mode, pool,
                         spark_base_args, job_args, conf)



import logging
import os
import sys
from contextlib import contextmanager

from pyspark import SparkContext, SparkConf
from pyspark.sql import SparkSession

from spyt.dependency_utils import require_yt_client
require_yt_client()

from yt.wrapper import YtClient  # noqa: E402
from yt.wrapper.http_helpers import get_token, get_user_name, get_proxy_address_netloc  # noqa: E402

from .arcadia import checked_extract_spark  # noqa: E402
from .utils import default_token, default_discovery_dir, get_spark_master, set_conf, \
    SparkDiscovery, parse_memory, format_memory, base_spark_conf, parse_bool, get_spyt_conf_dir, \
    check_spark_version  # noqa: E402
from .conf import read_remote_conf, read_global_conf, validate_versions_compatibility, \
    read_cluster_conf, SELF_VERSION  # noqa: E402


logger = logging.getLogger(__name__)


class Defaults(object):
    LOCAL_CONF_PATH = os.path.join(os.getenv("HOME"), "spyt.yaml") if os.getenv("HOME") else None


class SparkInfo(object):
    def __init__(self, spark):
        self.spark = spark
        pass

    def _repr_html_(self):
        yt_proxy = self.spark.conf.get("spark.hadoop.yt.proxy")
        client = yt_client(self.spark)
        discovery = SparkDiscovery()
        shs_url = SparkDiscovery.getOption(discovery.shs(), client=client)
        app_id = self.spark.conf.get("spark.app.id")
        master_web_ui = SparkDiscovery.get(discovery.master_webui(), client=client)
        spyt_version = self.spark.conf.get("spark.yt.version")
        spyt_cluster_version = self.spark.conf.get("spark.yt.cluster.version")
        return """
            <div>
                <p><b>SPYT</b></p>

                <p><a href="http://{master_web_ui}">Master Web UI</a></p>
                <p><a href="http://{shs_url}/history/{app_id}/jobs/">Spark history server</a></p>
                <dl>
                  <dt>Yt Cluster</dt>
                    <dd><code>{yt_proxy}</code></dd>
                  <dt>SPYT Cluster version</dt>
                    <dd><code>{spyt_cluster_version}</code></dd>
                  <dt>SPYT Library version</dt>
                    <dd><code>{spyt_version}</code></dd>
                </dl>

                {sc_HTML}
            </div>
        """.format(
            yt_proxy=yt_proxy,
            master_web_ui=master_web_ui,
            spyt_cluster_version=spyt_cluster_version,
            spyt_version=spyt_version,
            shs_url=shs_url,
            app_id=app_id,
            sc_HTML=self.spark.sparkContext._repr_html_()
        )


def info(spark):
    return SparkInfo(spark)


@contextmanager
def spark_session(num_executors=None,
                  yt_proxy=None,
                  discovery_path=None,
                  app_name=None,
                  cores_per_executor=None,
                  executor_memory_per_core=None,
                  driver_memory=None,
                  dynamic_allocation=False,
                  spark_conf_args=None,
                  local_conf_path=Defaults.LOCAL_CONF_PATH,
                  client=None,
                  spyt_version=None):
    def do_create_inner_cluster_session():
        return connect(
            num_executors=num_executors,
            yt_proxy=yt_proxy,
            discovery_path=discovery_path,
            app_name=app_name,
            cores_per_executor=cores_per_executor,
            executor_memory_per_core=executor_memory_per_core,
            driver_memory=driver_memory,
            dynamic_allocation=dynamic_allocation,
            spark_conf_args=spark_conf_args,
            local_conf_path=local_conf_path,
            client=client,
            spyt_version=spyt_version,
        )

    with _create_spark_session(do_create_inner_cluster_session) as spark:
        yield spark


@contextmanager
def direct_spark_session(yt_proxy, conf: SparkConf = None):
    with _create_spark_session(lambda: connect_direct(yt_proxy, conf)) as spark:
        yield spark


@contextmanager
def _create_spark_session(do_create_spark_session):
    spark_session_already_existed = _spark_session_exists()
    spark = do_create_spark_session()
    exception = None
    try:
        yield spark
    except Exception as e:
        exception = e
    finally:
        if spark_session_already_existed:
            _raise_first(exception)
        else:
            stop(spark, exception)


def get_spark_discovery(discovery_path, conf, client=None):
    discovery_path = discovery_path or conf.get("discovery_path") or conf.get(
        "discovery_dir") or default_discovery_dir(client=client)
    return SparkDiscovery(discovery_path=discovery_path)


def create_yt_client(yt_proxy, conf):
    yt_proxy = yt_proxy or conf.get("yt_proxy", os.getenv("YT_PROXY"))
    yt_token = conf.get("yt_token") or default_token()
    return YtClient(proxy=yt_proxy, token=yt_token)


def create_yt_client_spark_conf(yt_proxy, spark_conf):
    yt_proxy = yt_proxy or spark_conf.get("spark.hadoop.yt.proxy") or os.getenv("SPARK_YT_PROXY") or os.getenv("SPARK_HADOOP_YT_PROXY")
    yt_token = spark_conf.get("spark.hadoop.yt.token") or os.getenv("SPARK_YT_TOKEN") or os.getenv("SPARK_HADOOP_YT_TOKEN")
    return YtClient(proxy=yt_proxy, token=yt_token)


def _configure_client_mode(spark_conf,
                           discovery_path,
                           local_conf,
                           client=None,
                           spyt_version=None):
    discovery = get_spark_discovery(discovery_path, local_conf, client=client)
    master = get_spark_master(discovery, rest=False, yt_client=client)
    set_conf(spark_conf, base_spark_conf(client=client, discovery=discovery))
    spark_conf.set("spark.master", master)
    os.environ["SPARK_YT_TOKEN"] = get_token(client=client)
    os.environ["SPARK_BASE_DISCOVERY_PATH"] = str(discovery.base_discovery_path)
    _set_spark_conf_dir()

    spyt_version = spyt_version or SELF_VERSION
    spark_cluster_version = spark_conf.get("spark.yt.cluster.version")
    validate_versions_compatibility(spyt_version, spark_cluster_version)
    spark_conf.set("spark.yt.version", spyt_version)


def _set_spark_conf_dir():
    if "SPARK_CONF_DIR" not in os.environ:
        os.environ["SPARK_CONF_DIR"] = get_spyt_conf_dir()


def _validate_resources(num_executors, cores_per_executor, executor_memory_per_core):
    if num_executors and (not isinstance(num_executors, int) or num_executors < 1):
        raise AssertionError("Invalid number of executors")

    if cores_per_executor and (not isinstance(cores_per_executor, int) or cores_per_executor < 1):
        raise AssertionError("Invalid number of cores per executor")

    if executor_memory_per_core and executor_memory_per_core < 1:
        raise AssertionError("Invalid amount of memory per core")


def _set_none_safe(conf, key, value):
    if value:
        conf.set(key, value)


class Environment(object):
    IS_CLUSTER_PYTHON_PATH = False

    @staticmethod
    def configure_python_path():
        os.environ["PYSPARK_PYTHON"] = "python%d.%d" % sys.version_info[:2]
        Environment.IS_CLUSTER_PYTHON_PATH = True

    @staticmethod
    def unset_python_path():
        if Environment.IS_CLUSTER_PYTHON_PATH:
            del (os.environ["PYSPARK_PYTHON"])
            Environment.IS_CLUSTER_PYTHON_PATH = False


def _read_local_conf(conf_path):
    import yaml
    if conf_path and os.path.isfile(conf_path):
        with open(conf_path) as f:
            return yaml.load(f, Loader=yaml.BaseLoader)
    return {}


def _configure_resources(spark_conf, local_conf,
                         num_executors,
                         cores_per_executor,
                         executor_memory_per_core,
                         driver_memory,
                         dynamic_allocation):
    num_executors = num_executors or local_conf.get("num_executors")
    cores_per_executor = cores_per_executor or local_conf.get("cores_per_executor")
    executor_memory_per_core = parse_memory(executor_memory_per_core or local_conf.get("executor_memory_per_core"))
    _validate_resources(num_executors, cores_per_executor, executor_memory_per_core)
    driver_memory = parse_memory(driver_memory or local_conf.get("driver_memory"))
    executor_memory = executor_memory_per_core * cores_per_executor if executor_memory_per_core and cores_per_executor else None
    max_cores = num_executors * cores_per_executor if num_executors and cores_per_executor else None

    _set_none_safe(spark_conf, "spark.dynamicAllocation.enabled", dynamic_allocation)
    _set_none_safe(spark_conf, "spark.dynamicAllocation.executorIdleTimeout", "10m")
    _set_none_safe(spark_conf, "spark.dynamicAllocation.cachedExecutorIdleTimeout", "60m")
    _set_none_safe(spark_conf, "spark.cores.max", max_cores)
    _set_none_safe(spark_conf, "spark.dynamicAllocation.maxExecutors", num_executors)
    _set_none_safe(spark_conf, "spark.executor.cores", cores_per_executor)
    _set_none_safe(spark_conf, "spark.executor.memory", format_memory(executor_memory))
    _set_none_safe(spark_conf, "spark.driver.memory", format_memory(driver_memory))
    _set_none_safe(spark_conf, "spark.driver.maxResultSize", format_memory(driver_memory))


def _build_spark_conf(num_executors=None,
                      yt_proxy=None,
                      discovery_path=None,
                      app_name=None,
                      cores_per_executor=None,
                      executor_memory_per_core=None,
                      driver_memory=None,
                      dynamic_allocation=None,
                      spark_conf_args=None,
                      local_conf_path=None,
                      client=None,
                      spyt_version=None):
    is_client_mode = os.getenv("IS_SPARK_CLUSTER") is None
    local_conf = {}
    spark_conf = SparkConf()

    if is_client_mode:
        local_conf = _read_local_conf(local_conf_path)
        client = client or create_yt_client(yt_proxy, local_conf)
        app_name = app_name or "PySpark for {}".format(get_user_name(client=client) or os.getenv("USER"))
        _configure_client_mode(spark_conf, discovery_path, local_conf, client, spyt_version)
        spark_cluster_version = spark_conf.get("spark.yt.cluster.version")
        spark_cluster_conf_path = spark_conf.get("spark.yt.cluster.confPath")
    else:
        client = client or create_yt_client_spark_conf(yt_proxy, spark_conf)
        spark_cluster_version = os.getenv("SPARK_CLUSTER_VERSION") or os.getenv("SPYT_CLUSTER_VERSION")
        spark_cluster_conf_path = os.getenv("SPARK_YT_CLUSTER_CONF_PATH")

    _set_none_safe(spark_conf, "spark.app.name", app_name)
    _configure_resources(spark_conf, local_conf,
                         num_executors, cores_per_executor, executor_memory_per_core,
                         driver_memory, dynamic_allocation)

    global_conf = read_global_conf(client=client)
    remote_conf = read_remote_conf(global_conf, spark_cluster_version, client=client)
    set_conf(spark_conf, remote_conf["spark_conf"])

    if is_client_mode:
        Environment.configure_python_path()

    spark_cluster_conf = read_cluster_conf(spark_cluster_conf_path, client=client).get("spark_conf") or {}
    spark_conf.setAll(spark_cluster_conf.items())
    if spark_conf_args is not None:
        spark_conf.setAll(spark_conf_args.items())
    enabler_map = remote_conf["enablers"]
    for key, value in enabler_map.items():
        if spark_conf.contains(key):
            new_value = parse_bool(spark_conf.get(key)) and parse_bool(str(value))
            spark_conf.set(key, str(new_value))

    ipv6_preference_enabled = parse_bool(spark_conf.get('spark.hadoop.yt.preferenceIpv6.enabled'))
    if ipv6_preference_enabled:
        spark_conf.set('spark.driver.extraJavaOptions', '-Djava.net.preferIPv6Addresses=true')
        if check_spark_version(less_than="3.4.0"):
            # Starting from spark 3.4.0 IPv6 preference for executors is dependent on driver
            spark_conf.set('spark.executor.extraJavaOptions', '-Djava.net.preferIPv6Addresses=true')

    return spark_conf


def connect(num_executors=5,
            yt_proxy=None,
            discovery_path=None,
            app_name=None,
            cores_per_executor=4,
            executor_memory_per_core="4G",
            driver_memory="1G",
            dynamic_allocation=True,
            spark_conf_args=None,
            local_conf_path=Defaults.LOCAL_CONF_PATH,
            client=None,
            spyt_version=None):
    conf = _build_spark_conf(
        num_executors=num_executors,
        yt_proxy=yt_proxy,
        discovery_path=discovery_path,
        app_name=app_name,
        cores_per_executor=cores_per_executor,
        executor_memory_per_core=executor_memory_per_core,
        driver_memory=driver_memory,
        dynamic_allocation=dynamic_allocation,
        spark_conf_args=spark_conf_args,
        local_conf_path=local_conf_path,
        client=client,
        spyt_version=spyt_version,
    )
    if _spark_session_exists():
        logger.warning("SparkSession already exists and will be reused. Some configurations may not be applied. "
                       "You can use spyt.stop(spark) method to close previous session")

    checked_extract_spark()

    spark = SparkSession.builder.config(conf=conf).getOrCreate()

    logger.info("SPYT Cluster version: {}".format(spark.conf.get("spark.yt.cluster.version", default=None)))
    logger.info("SPYT library version: {}".format(spark.conf.get("spark.yt.version", default=None)))
    shs_url = _shs_url(discovery_path, spark)
    if shs_url is not None:
        logger.info("SHS link: {}".format(shs_url))

    # backward compatibility of SPYT-48
    if spark.conf.get("spark.sql.files.maxPartitionBytes") == "5000000":
        spark.conf.set("spark.sql.files.maxPartitionBytes", "2Gb")

    return spark


def connect_direct(yt_proxy: str = None, conf: SparkConf = None):
    checked_extract_spark()
    _set_spark_conf_dir()

    conf = conf or SparkConf()
    if not conf.contains("spark.ytsaurus.python.version"):
        conf.set("spark.ytsaurus.python.version", f"{sys.version_info.major}.{sys.version_info.minor}")
    if conf.contains("spark.executor.resource.gpu.amount"):
        conf.set("spark.executor.resource.gpu.discoveryScript", "./spyt-package/bin/getGpusResources.sh")
    spark = SparkSession.builder.config(conf=conf).master(
        "ytsaurus://" + (yt_proxy or get_proxy_address_netloc())
    ).getOrCreate()
    return spark


def _shs_url(discovery_path, spark):
    discovery = SparkDiscovery(discovery_path=discovery_path)
    if discovery.base_discovery_path is not None:
        client = yt_client(spark)
        shs_url = SparkDiscovery.getOption(discovery.shs(), client=client)
        if shs_url is not None:
            app_id = spark.conf.get("spark.app.id")
            return "http://{}/history/{}/jobs/".format(shs_url, app_id)


class CachedPy4JError(Exception):
    def __init__(self, py4j_error):
        self.cached_str = str(py4j_error)

    def __str__(self):
        return self.cached_str


def cache_exception(exc):
    if exc is None:
        return None
    else:
        cached = CachedPy4JError(exc)
        cached.with_traceback(exc.__traceback__)
        return cached


def stop(spark, exception=None):
    is_client_mode = spark.conf.get("spark.submit.deployMode") == "client"
    if exception is not None:
        logger.error("Shutdown SparkSession after exception: {}".format(exception))
    exception_c = cache_exception(exception)

    def stop_fault_handler(e1):
        return _try_with_safe_finally(
            lambda: _shutdown_jvm(spark) if is_client_mode else None,
            lambda e2: shutdown_jfv_fault_handler(e1, e2)
        )

    def shutdown_jfv_fault_handler(e1, e2):
        return _try_with_safe_finally(
            lambda: Environment.unset_python_path(),
            lambda e3: get_first_exception(exception_c, e1, e2, e3)
        )

    def get_first_exception(*exceptions):
        for e in exceptions:
            if e:
                return e
        return None

    main_exception = _try_with_safe_finally(
        lambda: spark.stop(),
        lambda e1: stop_fault_handler(e1)
    )
    if main_exception:
        raise main_exception


def is_stopped(spark):
    spark._jsc.sc().isStopped()


def yt_client(spark):
    yt_proxy = spark.conf.get("spark.hadoop.yt.proxy")
    yt_token = spark.conf.get("spark.hadoop.yt.token")
    return YtClient(proxy=yt_proxy, token=yt_token)


def jvm_process_pid():
    return SparkContext._gateway.proc.pid


def _close_yt_client(spark):
    spark._jvm.tech.ytsaurus.spyt.fs.YtClientProvider.close()


def _try_with_safe_finally(try_func, finally_func):
    exception = None
    try:
        try_func()
    except Exception as e:
        logger.error("Unexpected error {}".format(e.message))
        exception = e
    return finally_func(cache_exception(exception))


def _raise_first(*exceptions):
    exception = None
    for e in exceptions:
        exception = exception or e
    if exception:
        raise exception


def _shutdown_jvm(spark):
    from subprocess import Popen
    proc = SparkContext._gateway.proc
    if not isinstance(proc, Popen):
        logger.warning("SparkSession cannot be closed properly, please update ytsaurus-spyt and Spark cluster")
        return
    proc.stdin.close()
    SparkContext._gateway.shutdown()
    proc.wait(timeout=15)
    SparkContext._gateway = None
    SparkContext._jvm = None
    spark._jvm = None
    SparkSession.builder._options = {}


def _spark_session_exists():
    return SparkSession._instantiatedSession is not None



import io
from datetime import datetime, timedelta, timezone
from typing import Union

import yt.yson as yt_yson
from pyspark import SparkContext
from pyspark.sql.column import _to_java_column, Column
from pyspark.sql.types import UserDefinedType, BinaryType, IntegralType, LongType, IntegerType

MIN_DATE32 = -53375809
MAX_DATE32 = 53375807
MIN_DATETIME64 = -4611669897600
MAX_DATETIME64 = 4611669811199
MIN_TIMESTAMP64 = -4611669897600000000
MAX_TIMESTAMP64 = 4611669811199999999
MAX_INTERVAL64 = MAX_TIMESTAMP64 - MIN_TIMESTAMP64
MIN_INTERVAL64 = - MAX_INTERVAL64


class DatetimeType(UserDefinedType):
    def needConversion(self):
        return True

    def serialize(self, obj):
        dt = obj.value
        if dt is not None:
            if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
                tz_utc = timezone.utc
                dt = dt.replace(tzinfo=tz_utc)
            return int(dt.timestamp())
        return None

    def deserialize(self, ts):
        if ts is not None:
            return Datetime(datetime.fromtimestamp(ts, timezone.utc).replace(microsecond=0, tzinfo=None))

    @classmethod
    def typeName(cls):
        return "datetime"

    def simpleString(self):
        return 'datetime'

    @classmethod
    def sqlType(cls):
        return LongType()

    @classmethod
    def module(cls):
        return 'spyt.types'

    @classmethod
    def scalaUDT(cls):
        return 'org.apache.spark.sql.spyt.types.DatetimeType'


class Datetime:
    __UDT__ = DatetimeType()

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "Datetime(%s)" % self.value

    def __str__(self):
        return "(%s)" % self.value

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
            other.value == self.value

    @classmethod
    def from_seconds(cls, seconds):
        if seconds is None:
            return None
        dt_value = (datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(seconds=seconds)) \
            .astimezone(timezone.utc).replace(tzinfo=None)
        return cls(dt_value)


class Date32Type(UserDefinedType):

    def needConversion(self):
        return True

    def serialize(self, obj):
        return int(obj.value) if (obj.value is not None) else None

    def deserialize(self, d):
        if d is not None and MIN_DATE32 <= d <= MAX_DATE32:
            return Date32(d)

    @classmethod
    def typeName(cls):
        return "date32"

    def simpleString(self):
        return 'date32'

    @classmethod
    def sqlType(cls):
        return IntegerType()

    @classmethod
    def module(cls):
        return 'spyt.types'

    @classmethod
    def scalaUDT(cls):
        return 'org.apache.spark.sql.spyt.types.Date32Type'


class Date32:
    __UDT__ = Date32Type()

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "Date32(%s)" % self.value

    def __str__(self):
        return "(%s)" % self.value

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
            other.value == self.value


class Datetime64Type(UserDefinedType):

    def needConversion(self):
        return True

    def serialize(self, obj):
        return int(obj.value) if (obj.value is not None) else None

    def deserialize(self, d):
        if d is not None and MIN_DATETIME64 <= d <= MAX_DATETIME64:
            return Datetime64(d)

    @classmethod
    def typeName(cls):
        return "datetime64"

    def simpleString(self):
        return 'datetime64'

    @classmethod
    def sqlType(cls):
        return LongType()

    @classmethod
    def module(cls):
        return 'spyt.types'

    @classmethod
    def scalaUDT(cls):
        return 'org.apache.spark.sql.spyt.types.Datetime64Type'


class Datetime64:
    __UDT__ = Datetime64Type()

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "Datetime64(%s)" % self.value

    def __str__(self):
        return "(%s)" % self.value

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
            other.value == self.value


class Timestamp64Type(UserDefinedType):

    def needConversion(self):
        return True

    def serialize(self, obj):
        return int(obj.value) if (obj.value is not None) else None

    def deserialize(self, d):
        if d is not None and MIN_TIMESTAMP64 <= d <= MAX_TIMESTAMP64:
            return Timestamp64(d)

    @classmethod
    def typeName(cls):
        return "timestamp64"

    def simpleString(self):
        return 'timestamp64'

    @classmethod
    def sqlType(cls):
        return LongType()

    @classmethod
    def module(cls):
        return 'spyt.types'

    @classmethod
    def scalaUDT(cls):
        return 'org.apache.spark.sql.spyt.types.Timestamp64Type'


class Timestamp64:
    __UDT__ = Timestamp64Type()

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "Timestamp64(%s)" % self.value

    def __str__(self):
        return "(%s)" % self.value

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
            other.value == self.value


class Interval64Type(UserDefinedType):

    def needConversion(self):
        return True

    def serialize(self, obj):
        return int(obj.value) if (obj.value is not None) else None

    def deserialize(self, d):
        if d is not None and MIN_INTERVAL64 <= d <= MAX_INTERVAL64:
            return Interval64(d)

    @classmethod
    def typeName(cls):
        return "interval64"

    def simpleString(self):
        return 'interval64'

    @classmethod
    def sqlType(cls):
        return LongType()

    @classmethod
    def module(cls):
        return 'spyt.types'

    @classmethod
    def scalaUDT(cls):
        return 'org.apache.spark.sql.spyt.types.Interval64Type'


class Interval64:
    __UDT__ = Interval64Type()

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "Interval64(%s)" % self.value

    def __str__(self):
        return "(%s)" % self.value

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
            other.value == self.value


class YsonType(UserDefinedType):
    def needConversion(self):
        return True

    @classmethod
    def typeName(cls):
        return "yson"

    @classmethod
    def module(cls):
        return 'spyt.types'

    @classmethod
    def scalaUDT(cls):
        return 'org.apache.spark.sql.spyt.types.YsonType'

    def simpleString(self):
        return 'yson'

    @classmethod
    def sqlType(cls):
        return BinaryType()

    def serialize(self, obj):
        return yt_yson.dumps(obj.value, "binary")

    def fromInternal(self, binary_data: bytearray) -> Union[yt_yson.YsonMap, None]:
        if binary_data is None:
            return None
        yson: yt_yson.YsonMap = yt_yson.load(io.BytesIO(binary_data))
        return yson


class Yson:
    __UDT__ = YsonType()

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "Yson(%s)" % self.value

    def __str__(self):
        return "(%s)" % self.value

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
            other.value == self.value


UINT64_MAX = 0xffffffffffffffff


class UInt64Type(IntegralType):
    """Unsigned 64-bit integer type
    """

    def simpleString(self):
        return 'uint64'

    @classmethod
    def typeName(cls):
        return "uint64"

    def needConversion(self):
        return True

    def toInternal(self, py_integer):
        if py_integer is None:
            return None
        if py_integer < 0 or py_integer > UINT64_MAX:
            raise ValueError(f"object of UInt64Type out of range, got: {py_integer}")
        return py_integer if py_integer <= 0x7fffffffffffffff else py_integer - UINT64_MAX - 1

    def fromInternal(self, j_long):
        return j_long if j_long is None or j_long >= 0 else (j_long & UINT64_MAX)


def uint64_to_string(number):
    if number is None:
        return None
    else:
        # convert to unsigned value
        return str(number & UINT64_MAX)


def string_to_uint64(number):
    if number is None:
        return None
    else:
        return int(number)


def uint64_to_string_udf(s_col):
    sc = SparkContext._active_spark_context
    cols = sc._gateway.new_array(sc._jvm.Column, 1)
    cols[0] = _to_java_column(s_col)
    jc = sc._jvm.org.apache.spark.sql.spyt.types.UInt64Long.toStringUdf().apply(cols)
    return Column(jc)


def string_to_uint64_udf(s_col):
    sc = SparkContext._active_spark_context
    cols = sc._gateway.new_array(sc._jvm.Column, 1)
    cols[0] = _to_java_column(s_col)
    jc = sc._jvm.org.apache.spark.sql.spyt.types.UInt64Long.fromStringUdf().apply(cols)
    return Column(jc)


def xx_hash64_zero_seed_udf(*s_cols):
    sc = SparkContext._active_spark_context
    sz = len(s_cols)
    cols = sc._gateway.new_array(sc._jvm.Column, sz)
    for i in range(sz):
        cols[i] = _to_java_column(s_cols[i])
    jc = sc._jvm.tech.ytsaurus.spyt.common.utils.XxHash64ZeroSeed.xxHash64ZeroSeedUdf(cols)
    return Column(jc)


def register_xxHash64ZeroSeed(spark):
    sc = SparkContext._active_spark_context
    sc._jvm.tech.ytsaurus.spyt.common.utils.XxHash64ZeroSeed.registerFunction(spark._jsparkSession)


def cityhash_udf(*s_cols):
    sc = SparkContext._active_spark_context
    sz = len(s_cols)
    cols = sc._gateway.new_array(sc._jvm.Column, sz)
    for i in range(sz):
        cols[i] = _to_java_column(s_cols[i])
    jc = sc._jvm.tech.ytsaurus.spyt.common.utils.CityHash.cityHashUdf(cols)
    return Column(jc)


def register_cityHash(spark):
    sc = SparkContext._active_spark_context
    sc._jvm.tech.ytsaurus.spyt.common.utils.CityHash.registerFunction(spark._jsparkSession)


def tuple_type(element_types):
    """
    :param element_types: List[DataType]
    :return: StructType
    """
    from pyspark.sql.types import StructType, StructField
    struct_fields = [StructField("_{}".format(i + 1), element_type) for i, element_type in enumerate(element_types)]
    return StructType(struct_fields)


def variant_over_struct_type(elements):
    """
    :param elements: List[Tuple[str, DataType]]
    :return: StructType
    """
    from pyspark.sql.types import StructType, StructField
    struct_fields = [StructField("_v{}".format(element_name), element_type) for element_name, element_type in elements]
    return StructType(struct_fields)


def variant_over_tuple_type(element_types):
    """
    :param element_types: List[DataType]
    :return: StructType
    """
    elements = [("_{}".format(i + 1), element_type) for i, element_type in enumerate(element_types)]
    return variant_over_struct_type(elements)



#! /usr/bin/env python3
import json
import sys
import traceback


def extract_spark_component(sensor):
    spark_component = None
    if '_driver_' in sensor:
        spark_component = 'driver'
    elif '_executor_' in sensor:
        spark_component = 'executor'
    return spark_component


def extract_app_id_and_clean_sensor(sensor, spark_component, spark_component_label_missed=False):
    sensor_parts = sensor.split('_')
    spark_application_id = sensor_parts[1]
    length_to_delete = 3 if spark_component == 'driver' else 4
    if spark_component_label_missed:
        length_to_delete -= 1
    del sensor_parts[1:length_to_delete]
    clean_sensor = "_".join(sensor_parts)
    return clean_sensor, spark_application_id


def process_sensors(payload):
    if 'sensors' in  payload:
        sensors = payload['sensors']
        component = payload["commonLabels"]['component']
        for sensor in sensors:
            sensor_name = sensor["labels"]["sensor"]
            is_spark_component_label_missed = False
            spark_component = extract_spark_component(sensor_name)
            if spark_component is None and component in ('driver', 'executor'):
                # special case with missed executor tag in name we inherit from job task name
                is_spark_component_label_missed = True
                spark_component = component
            if spark_component:
                clean_name, app_id = extract_app_id_and_clean_sensor(sensor_name, spark_component,
                                                                     is_spark_component_label_missed)
                labels = sensor["labels"]
                labels['sensor'] = clean_name
                labels['spark_application_id'] = app_id
                labels['spark_component'] = spark_component


def process_line(line):
    try:
        message = json.loads(line)
        payload = json.loads(message['payload'])
        process_sensors(payload)
        message["payload"] = json.dumps(payload)
        json.dump(message, sys.stdout)
        sys.stdout.write('\n')
    except Exception as e:
        error_type = type(e).__name__
        error_msg = f"Error: {error_type}: {str(e)} in line: {line}\n"
        stack_trace = traceback.format_exc()
        error_msg += f"Stack trace:\n{stack_trace}\n"
        sys.stderr.write(error_msg)
        sys.stderr.flush()

        # keep initial message
        sys.stdout.write(line)


def main():
    while True:
        line = sys.stdin.readline().rstrip()
        if len(line) == 0:
            break
        process_line(line)
        sys.stdout.flush()


if __name__ == "__main__":
    main()



def require_yt_client():
    try:
        import yt.wrapper  # noqa: F401
        from yt.wrapper import YtClient  # noqa: F401
    except ImportError as e:
        raise ImportError(
            "Please install ytsaurus-client (or yandex-yt for internal Yandex users). "
            "These libraries cannot be installed at the same time"
        ) from e


def require_pyspark():
    compatible_versions_str = '>=3.2.2,<4.0.0'
    try:
        import pyspark
        try:
            from packaging.specifiers import SpecifierSet
            from .version import __version__ as spyt_version
            compatible_versions = SpecifierSet(compatible_versions_str)
            if pyspark.__version__ not in compatible_versions:
                raise AssertionError(
                    f"ytsaurus-spyt {spyt_version} is compatible with pyspark{compatible_versions_str}, "
                    f"however you have pyspark {pyspark.__version__}. Please install compatible pyspark package "
                    f"or try to upgrade ytsaurus-spyt"
                )
        except ImportError:
            pass
    except ImportError as e:
        raise ImportError(
            f"Please install pyspark module within versions range {compatible_versions_str}. "
            f"It is required to use spyt."
        ) from e



import logging
import os
import random
import re
import subprocess

from spyt.dependency_utils import require_yt_client

require_yt_client()

from yt.wrapper.cypress_commands import exists, copy as cypress_copy, get as yt_get  # noqa: E402
from yt.wrapper.acl_commands import check_permission  # noqa: E402
from yt.wrapper.file_commands import upload_file_to_cache  # noqa: E402
from yt.wrapper.http_helpers import get_token, get_user_name  # noqa: E402
from yt.wrapper.operation_commands import TimeWatcher, \
    abort_operation, get_operation_state  # noqa: E402
from yt.wrapper.run_operation_commands import run_operation  # noqa: E402

try:
    from yt.wrapper.operation_commands import process_operation_unsuccessful_finish_state
except Exception:
    # COMPAT(alex-shishkin): For ytsaurus-client <= 0.13.7
    from yt.wrapper.operation_commands \
        import process_operation_unsuccesful_finish_state as process_operation_unsuccessful_finish_state

from .conf import read_remote_conf, validate_cluster_version, \
    update_config_inplace, validate_custom_params, validate_mtn_config, \
    latest_ytserver_proxy_path, read_global_conf, \
    worker_num_limit, validate_worker_num, read_cluster_conf, validate_ssd_config, cuda_toolkit_version  # noqa: E402
from .utils import get_spark_master, base_spark_conf, SparkDiscovery, SparkCluster, call_get_proxy_address_url, \
    parse_bool, _add_conf, check_spark_version  # noqa: E402
from .enabler import SpytEnablers  # noqa: E402
from .spec import SparkDefaultArguments, CommonComponentConfig, MasterConfig, WorkerConfig, HistoryServerConfig, \
    LivyConfig, build_spark_operation_spec, WorkerResources  # noqa: E402
from .version import __scala_version__  # noqa: E402
from .submit import create_base_spark_env  # noqa: E402

logger = logging.getLogger(__name__)


def _add_master(discovery, spark_args, rest, client=None):
    spark_args.append("--master")
    spark_args.append(get_spark_master(discovery, rest=False, yt_client=client))
    _add_conf({
        "spark.rest.master": get_spark_master(discovery, rest=True, yt_client=client)
    }, spark_args)


def _add_shs_option(discovery, spark_args, client=None):
    shs = SparkDiscovery.getOption(discovery.shs(), client=client)
    if shs is not None:
        _add_conf({
            "spark.rest.shs": shs
        }, spark_args)


def _add_base_spark_conf(client, discovery, spark_args):
    _add_conf(base_spark_conf(client, discovery), spark_args)


def _add_job_args(job_args, spark_args):
    if job_args:
        for k, v in job_args.items():
            spark_args.append("--{}".format(k))
            spark_args.append(v)


def _create_spark_env(client, spark_home):
    spark_env = create_base_spark_env(spark_home)
    spark_env["SPARK_USER"] = get_user_name(client=client)
    spark_env["SPARK_YT_TOKEN"] = get_token(client=client)
    spark_env["SPARK_YT_PROXY"] = call_get_proxy_address_url(client=client)
    return spark_env


def _wait_op_start(op, operation_path=None, client=None):
    for state in op.get_state_monitor(TimeWatcher(5.0, 5.0, 0.0)):
        if state.is_running() and operation_path is not None and exists(operation_path, client=client):
            return op
        elif state.is_unsuccessfully_finished():
            process_operation_unsuccessful_finish_state(op, op.get_error(state))
        else:
            op.printer(state)


def _wait_child_start(op, spark_discovery, client):
    _wait_op_start(op, spark_discovery.children_operations().join(op.id), client)


def _wait_master_start(op, spark_discovery, client):
    _wait_op_start(op, spark_discovery.operation().join(op.id), client)


def _wait_livy_start(op, spark_discovery, client):
    _wait_op_start(op, spark_discovery.livy(), client)


def _jmx_opts(port):
    return "-Dcom.sun.management.jmxremote " \
           "-Dcom.sun.management.jmxremote.port={} " \
           "-Dcom.sun.management.jmxremote.authenticate=false " \
           "-Dcom.sun.management.jmxremote.ssl=false".format(port)


def _process_ipv6_preference(dynamic_config, enable_preference_ipv6, enablers):
    if (enable_preference_ipv6 is None and
            'spark.hadoop.yt.preferenceIpv6.enabled' in dynamic_config['spark_conf']):
        enable_preference_ipv6 = parse_bool(dynamic_config['spark_conf']['spark.hadoop.yt.preferenceIpv6.enabled'])
    if enable_preference_ipv6 is not None:
        enablers.enable_preference_ipv6 = enable_preference_ipv6


def submit(discovery_path, spark_home, deploy_mode, spark_conf, job_class, jar_path, job_args, client=None):
    spark_args = ["--deploy-mode", deploy_mode]

    _add_conf(spark_conf, spark_args)

    spark_args.append("--class")
    spark_args.append(job_class)
    spark_args.append(jar_path)

    _add_job_args(job_args, spark_args)

    raw_submit(discovery_path=discovery_path,
               spark_home=spark_home,
               spark_args=spark_args,
               client=client)


def submit_python(discovery_path, spark_home, deploy_mode, spark_conf, main_py_path, py_files, job_args, client=None):
    spark_args = ["--deploy-mode", deploy_mode]

    _add_conf(spark_conf, spark_args)

    spark_args.append("--py-files")
    spark_args.append(py_files)
    spark_args.append(main_py_path)

    _add_job_args(job_args, spark_args)

    raw_submit(discovery_path=discovery_path,
               spark_home=spark_home,
               spark_args=spark_args,
               client=client)


def raw_submit(discovery_path, spark_home, spark_args,
               python_version=None, local_files=True, client=None):
    spark_submit_path = "{}/bin/spark-submit".format(spark_home)
    spark_base_args = [spark_submit_path]
    permission_status = check_permission(user=client.get_user_name(),
                                         permission='read', path=discovery_path, client=client)
    if permission_status.get('action', 'deny') != 'allow':
        raise RuntimeError(
            'No permission for reading cluster, actual permission status is ' + str(permission_status))
    discovery = SparkDiscovery(discovery_path=discovery_path)

    cluster_conf = read_cluster_conf(str(discovery.conf()), client)
    spark_conf = cluster_conf['spark_conf']
    dedicated_driver_op = parse_bool(spark_conf.get('spark.dedicated_operation_mode'))
    ipv6_preference_enabled = parse_bool(spark_conf.get('spark.hadoop.yt.preferenceIpv6.enabled'))

    _add_master(discovery, spark_base_args, rest=True, client=client)
    _add_shs_option(discovery, spark_base_args, client=client)
    _add_base_spark_conf(client, discovery, spark_base_args)
    _add_python_version(python_version, spark_base_args)
    _add_dedicated_driver_op_conf(spark_base_args, dedicated_driver_op)
    _add_ipv6_preference(ipv6_preference_enabled, spark_base_args)
    spark_env = _create_spark_env(client, spark_home)

    if local_files:
        remote_paths = {}
        new_spark_args = []
        # Matches "FS:/..." path sequence
        local_files_pattern = "[A-Za-z0-9]+:\/[^,]+(,[A-Za-z0-9]+:\/[^,]+)*"  # noqa: W605
        for spark_arg in spark_args:
            is_file_list = re.fullmatch(local_files_pattern, spark_arg)
            if not is_file_list:
                new_spark_args.append(spark_arg)
                continue
            file_list = spark_arg.split(",")
            new_file_list = []
            for single_file in file_list:
                if single_file.startswith('local:/'):
                    local_file_path = single_file[7:]  # Drops prefix
                    if local_file_path not in remote_paths:
                        _, file_extension = os.path.splitext(local_file_path)
                        destination = upload_file_to_cache(local_file_path, client=client)
                        destination_ext = "{}{}".format(destination, file_extension)
                        if file_extension != "":
                            cypress_copy(destination, destination_ext, ignore_existing=True, client=client)  # Extension is necessary
                        else:
                            logger.warn("%s has no extension. Usually such files are unacceptable")
                        logger.info("%s has been uploaded to YT as %s", local_file_path, destination_ext)
                        remote_paths[local_file_path] = "yt:/{}".format(destination_ext)
                    new_file_list.append(remote_paths[local_file_path])
                else:
                    new_file_list.append(single_file)
            new_spark_arg = ",".join(new_file_list)
            new_spark_args.append(new_spark_arg)
        spark_args = new_spark_args

    # replace stdin to avoid https://bugs.openjdk.java.net/browse/JDK-8211842
    return subprocess.call(spark_base_args + spark_args, env=spark_env, stdin=subprocess.PIPE)


def _add_dedicated_driver_op_conf(spark_args, dedicated_driver_op):
    if dedicated_driver_op:
        _add_conf({
            "spark.driver.resource.driverop.amount": 1
        }, spark_args)


def _add_python_version(python_version, spark_args):
    if python_version is not None:
        _add_conf({"spark.pyspark.python": f"python{python_version}"}, spark_args)


def _add_ipv6_preference(ipv6_preference_enabled, spark_args):
    if ipv6_preference_enabled:
        ipv6_conf = {'spark.driver.extraJavaOptions': '-Djava.net.preferIPv6Addresses=true'}
        if check_spark_version(less_than="3.4.0"):
            # Starting from spark 3.4.0 IPv6 preference for executors is dependent on driver
            ipv6_conf['spark.executor.extraJavaOptions'] = '-Djava.net.preferIPv6Addresses=true'
        _add_conf(ipv6_conf, spark_args)


def wrap_cached_jar(path, jar_caching_enabled):
    if jar_caching_enabled and path.startswith("yt:/"):
        return "ytCached:/" + path[4:]
    else:
        return path


def shell(discovery_path, spark_home, spark_args, spyt_version=None, client=None):
    spark_shell_path = "{}/bin/spark-shell".format(spark_home)
    spark_base_args = [spark_shell_path]
    discovery = SparkDiscovery(discovery_path=discovery_path)

    spark_conf = read_cluster_conf(str(discovery.conf()), client)['spark_conf']
    ipv6_preference_enabled = parse_bool(spark_conf.get('spark.hadoop.yt.preferenceIpv6.enabled'))

    _add_master(discovery, spark_base_args, rest=False, client=client)
    _add_shs_option(discovery, spark_base_args, client=client)
    _add_base_spark_conf(client, discovery, spark_base_args)
    _add_conf({"spark.ui.showConsoleProgress": "true"}, spark_base_args)
    _add_conf(spark_conf, spark_base_args)
    _add_ipv6_preference(ipv6_preference_enabled, spark_base_args)
    spark_env = _create_spark_env(client, spark_home)

    os.execve(spark_shell_path, spark_base_args + spark_args, spark_env)


def stop_spark_cluster(discovery_path, client):
    """Stop Spark cluster
    :param discovery_path: Cypress path for discovery files and logs
    :param client: YtClient
    """
    spark_discovery = SparkDiscovery(discovery_path=discovery_path)
    abort_spark_operations(spark_discovery, client)


def abort_spark_operations(spark_discovery, client):
    current_operation_id = SparkDiscovery.getOption(
        spark_discovery.operation(), client=client)
    error = None
    if current_operation_id is not None and get_operation_state(
            current_operation_id, client=client).is_running():
        error = abort_operation_silently(current_operation_id, client=client)
        for child_id in SparkDiscovery.getOptions(spark_discovery.children_operations(), client):
            if get_operation_state(child_id, client=client).is_running():
                abort_operation_silently(child_id, client=client)
    if error:
        raise error


def get_base_cluster_config(global_conf, spark_cluster_version, params, base_discovery_path=None, client=None):
    dynamic_config = SparkDefaultArguments.get_params()
    update_config_inplace(dynamic_config, read_remote_conf(global_conf, spark_cluster_version, client=client))
    update_config_inplace(dynamic_config, params)
    if base_discovery_path is not None:
        dynamic_config['spark_conf']['spark.base.discovery.path'] = base_discovery_path
    return dynamic_config


def run_operation_wrapper(op_builder, address_path, client):
    op = None
    try:
        op = run_operation(op_builder, sync=False, client=client)
        if address_path is not None:
            _wait_op_start(op, address_path, client)
            address = SparkDiscovery.get(address_path, client=client)
            logger.info("URL: http://{0}".format(address))
        logger.info("Operation %s", op.id)
    except Exception as err:
        logging.error(err, exc_info=True)
        abort_operation_silently(op, client)
    return op


def start_livy_server(operation_alias=None, discovery_path=None, pool=None, enable_tmpfs=False, network_project=None,
                      tvm_id=None, tvm_secret=None, params=None, spark_cluster_version=None, enablers=None,
                      enable_preference_ipv6=None, client=None,
                      preemption_mode="normal", cluster_log_level="INFO",
                      livy_driver_cores=SparkDefaultArguments.LIVY_DRIVER_CORES,
                      livy_driver_memory=SparkDefaultArguments.LIVY_DRIVER_MEMORY,
                      livy_max_sessions=SparkDefaultArguments.LIVY_MAX_SESSIONS, spark_master_address=None,
                      rpc_job_proxy=False, rpc_job_proxy_thread_pool_size=4, tcp_proxy_range_start=30000,
                      tcp_proxy_range_size=100, enable_stderr_table=False, master_group_id=None, group_id=None,
                      cluster_java_home=None):
    if discovery_path is None and group_id is None:
        raise RuntimeError("Either discovery path or discovery group id must be provided")

    enablers = enablers or SpytEnablers()

    if spark_master_address is None:
        logger.warning("Spark master address is not specified, "
                       "standalone cluster will be discovered. "
                       "If you want use direct submit, "
                       "please provide the option `--spark-master-address ytsaurus://<ytsaurus-proxy>`")

    if spark_cluster_version is None:
        spark_cluster_version = __scala_version__

    validate_cluster_version(spark_cluster_version, client=client)
    validate_custom_params(params)
    validate_mtn_config(enablers, network_project, tvm_id, tvm_secret)

    global_conf = read_global_conf(client=client)
    dynamic_config = get_base_cluster_config(global_conf, spark_cluster_version, params, discovery_path, client)
    _process_ipv6_preference(dynamic_config, enable_preference_ipv6, enablers)
    enablers.apply_config(dynamic_config)

    spark_discovery = None
    if discovery_path is not None:
        spark_discovery = SparkDiscovery(discovery_path=discovery_path)
        spark_discovery.create(client)
        spark_discovery.set_cluster_version_if_none(spark_cluster_version, client)

    common_config = CommonComponentConfig(
        operation_alias, pool, enable_tmpfs, network_project, tvm_id, tvm_secret, enablers, preemption_mode,
        cluster_log_level, rpc_job_proxy, rpc_job_proxy_thread_pool_size, False, tcp_proxy_range_start,
        tcp_proxy_range_size, enable_stderr_table, "livy", spark_discovery, group_id, cluster_java_home
    )
    livy_config = LivyConfig(
        livy_driver_cores, livy_driver_memory, livy_max_sessions, spark_master_address,
        master_group_id, network_project
    )
    livy_builder = build_spark_operation_spec(config=dynamic_config, client=client, job_types=['livy'],
                                              common_config=common_config, livy_config=livy_config)
    address_path = spark_discovery.livy() if spark_discovery is not None else None
    return run_operation_wrapper(livy_builder, address_path, client)


def start_history_server(operation_alias=None, discovery_path=None, pool=None, enable_tmpfs=False,
                         history_server_memory_limit=SparkDefaultArguments.SPARK_HISTORY_SERVER_MEMORY_LIMIT,
                         history_server_cpu_limit=SparkDefaultArguments.SPARK_HISTORY_SERVER_CPU_LIMIT,
                         history_server_memory_overhead=SparkDefaultArguments.SPARK_HISTORY_SERVER_MEMORY_OVERHEAD,
                         network_project=None, tvm_id=None, tvm_secret=None, advanced_event_log=True,
                         params=None, shs_location=None, spark_cluster_version=None, enablers=None,
                         enable_preference_ipv6=None, client=None,
                         preemption_mode="normal", cluster_log_level="INFO", rpc_job_proxy=False,
                         rpc_job_proxy_thread_pool_size=4, tcp_proxy_range_start=30000,
                         tcp_proxy_range_size=100, enable_stderr_table=False, cluster_java_home=None):
    enablers = enablers or SpytEnablers()

    spark_discovery = SparkDiscovery(discovery_path=discovery_path)

    global_conf = read_global_conf(client=client)
    if spark_cluster_version is None:
        spark_cluster_version = __scala_version__

    validate_cluster_version(spark_cluster_version, client=client)
    validate_custom_params(params)
    validate_mtn_config(enablers, network_project, tvm_id, tvm_secret)

    dynamic_config = get_base_cluster_config(global_conf, spark_cluster_version, params,
                                             spark_discovery.base_discovery_path, client)
    _process_ipv6_preference(dynamic_config, enable_preference_ipv6, enablers)
    enablers.apply_config(dynamic_config)

    spark_discovery.create(client)

    common_config = CommonComponentConfig(
        operation_alias, pool, enable_tmpfs, network_project, tvm_id, tvm_secret, enablers, preemption_mode,
        cluster_log_level, rpc_job_proxy, rpc_job_proxy_thread_pool_size, False, tcp_proxy_range_start,
        tcp_proxy_range_size, enable_stderr_table, "shs", spark_discovery, None, cluster_java_home
    )
    hs_config = HistoryServerConfig(
        history_server_memory_limit, history_server_cpu_limit, history_server_memory_overhead, shs_location,
        advanced_event_log
    )
    hs_builder = build_spark_operation_spec(config=dynamic_config, client=client, job_types=['history'],
                                            common_config=common_config, hs_config=hs_config)
    return run_operation_wrapper(hs_builder, spark_discovery.shs(), client)


def set_random_shuffle_service_port(spark_conf):
    if 'spark.shuffle.service.port' not in spark_conf:
        start_port = int(spark_conf.get('spark.shuffle.service.port.interval.start', "27050"))
        interval_size = int(spark_conf.get('spark.shuffle.service.port.interval.size', "50"))
        spark_conf['spark.shuffle.service.port'] = str(start_port + random.randint(0, interval_size - 1))


def start_spark_cluster(worker_cores, worker_memory, worker_num, worker_cores_overhead=None,
                        worker_memory_overhead=SparkDefaultArguments.SPARK_WORKER_MEMORY_OVERHEAD,
                        worker_timeout=SparkDefaultArguments.SPARK_WORKER_TIMEOUT,
                        operation_alias=None, discovery_path=None, pool=None,
                        enable_tmpfs=True, tmpfs_limit=SparkDefaultArguments.SPARK_WORKER_TMPFS_LIMIT,
                        ssd_limit=SparkDefaultArguments.SPARK_WORKER_SSD_LIMIT,
                        ssd_account=None, worker_disk_name="default", worker_disk_limit=None, worker_disk_account=None,
                        worker_port=SparkDefaultArguments.SPARK_WORKER_PORT,
                        master_memory_limit=SparkDefaultArguments.SPARK_MASTER_MEMORY_LIMIT,
                        master_port=SparkDefaultArguments.SPARK_MASTER_PORT,
                        enable_history_server=True,
                        history_server_memory_limit=SparkDefaultArguments.SPARK_HISTORY_SERVER_MEMORY_LIMIT,
                        history_server_cpu_limit=SparkDefaultArguments.SPARK_HISTORY_SERVER_CPU_LIMIT,
                        history_server_memory_overhead=SparkDefaultArguments.SPARK_HISTORY_SERVER_MEMORY_OVERHEAD,
                        network_project=None, abort_existing=False, tvm_id=None, tvm_secret=None,
                        advanced_event_log=False, worker_log_transfer=False, worker_log_json_mode=False,
                        worker_log_update_interval=SparkDefaultArguments.SPARK_WORKER_LOG_UPDATE_INTERVAL,
                        worker_log_table_ttl=SparkDefaultArguments.SPARK_WORKER_LOG_TABLE_TTL,
                        params=None, shs_location=None, spark_cluster_version=None, enablers=None,
                        enable_preference_ipv6=None, client=None,
                        preemption_mode="normal", cluster_log_level="INFO", enable_multi_operation_mode=False,
                        dedicated_operation_mode=False, driver_cores=None, driver_memory=None, driver_num=None,
                        driver_cores_overhead=None, driver_timeout=None, autoscaler_period=None,
                        autoscaler_metrics_port=None, autoscaler_sliding_window=None,
                        autoscaler_max_free_workers=None, autoscaler_slot_increment_step=None,
                        enable_livy=False, livy_driver_cores=SparkDefaultArguments.LIVY_DRIVER_CORES,
                        livy_driver_memory=SparkDefaultArguments.LIVY_DRIVER_MEMORY,
                        livy_max_sessions=SparkDefaultArguments.LIVY_MAX_SESSIONS, rpc_job_proxy=False,
                        rpc_job_proxy_thread_pool_size=4, tcp_proxy_range_start=30000,
                        tcp_proxy_range_size=100, enable_stderr_table=False, group_id=None,
                        worker_gpu_limit=0, cluster_java_home=None, enable_ytsaurus_shuffle=False):
    """Start Spark cluster
    :param operation_alias: alias for the underlying YT operation
    :param pool: pool for the underlying YT operation
    :param discovery_path: Cypress path for discovery files and logs
    :param worker_cores: number of cores that will be available on worker
    :param worker_memory: amount of memory that will be available on worker
    :param worker_num: number of workers
    :param worker_cores_overhead: additional worker cores
    :param worker_memory_overhead: additional worker memory
    :param worker_timeout: timeout to fail master waiting
    :param worker_port: starting port to bind worker rpc endpoint
    :param enable_tmpfs: mounting ram memory as directory 'tmpfs'
    :param tmpfs_limit: limit of tmpfs usage, default 150G
    :param ssd_limit: limit of ssd usage, default None, ssd disabled
    :param ssd_account: account for ssd quota
    :param worker_disk_name: medium name
    :param worker_disk_limit: limit of disk usage, default None, disk disabled
    :param worker_disk_account: account for disk quota
    :param master_memory_limit: memory limit for master, default 2G
    :param master_port: starting port to bind master rpc endpoint
    :param enable_history_server: enables SHS
    :param history_server_memory_limit: memory limit for history server, default 16G,
    total memory for SHS job is history_server_memory_limit + history_server_memory_overhead
    :param history_server_cpu_limit: cpu limit for history server, default 20
    :param history_server_memory_overhead: memory overhead for history server, default 2G,
    total memory for SHS job is history_server_memory_limit + history_server_memory_overhead
    :param spark_cluster_version: SPYT cluster version
    :param network_project: YT network project
    :param abort_existing: abort existing running operation
    :param advanced_event_log: advanced log format for history server (requires dynamic tables write permission)
    :param worker_log_transfer: sending logs from workers to yt
    :param worker_log_json_mode: using json for worker logs
    :param worker_log_update_interval: intervals between log updates
    :param worker_log_table_ttl: TTL of yt table with worker logs
    :param tvm_id: TVM id for network project
    :param tvm_secret: TVM secret for network project
    :param params: YT operation params: file_paths, layer_paths, operation_spec, environment, spark_conf
    :param shs_location: hard set path to log directory
    :param enablers: ...
    :param enable_preference_ipv6: Whether to prefer ipv6 addresses or ipv4, default is release config setting
    :param client: YtClient
    :param preemption_mode: 'normal' or 'graceful' for graceful preemption
    :param cluster_log_level: level for cluster logs
    :param enable_multi_operation_mode: use several vanilla operations for one cluster
    :param dedicated_operation_mode: use dedicated operation for drivers
    :param driver_cores: number of cores that will be available on driver worker
    :param driver_memory: amount of memory that will be available on driver worker
    :param driver_num: number of driver workers
    :param driver_cores_overhead: additional driver worker cores
    :param driver_timeout: timeout to fail master waiting
    :param autoscaler_period: time between autoscaler calls in scala duration string format
    :param autoscaler_metrics_port: port for exposing of autoscaler metrics (deprecated, not used)
    :param autoscaler_sliding_window: size of autoscaler actions sliding window (in number of action) to downscale
    :param autoscaler_max_free_workers: maximum number of free workers
    :param autoscaler_slot_increment_step: autoscaler workers increment step
    :param enable_livy: start livy server
    :param livy_driver_cores: core limit for livy drivers
    :param livy_driver_memory: memory limit for livy drivers
    :param livy_max_sessions: session count limit for livy server
    :param rpc_job_proxy: using RPC proxy in job proxy
    :param rpc_job_proxy_thread_pool_size: RPC proxy thread pool size
    :param tcp_proxy_range_start: start port of TCP proxy allocation range
    :param tcp_proxy_range_size: size of TCP proxy allocation range
    :param enable_stderr_table: enables writing YT operation logs to stderr table
    :param group_id: discovery group id
    :param worker_gpu_limit: number of gpu for each worker
    :param cluster_java_home: custom java home for cluster vanilla operation
    :param enable_ytsaurus_shuffle: use YTsaurus shuffle service
    :return:
    """
    worker_res = WorkerResources(worker_cores, worker_memory, worker_num, worker_cores_overhead, worker_timeout,
                                 worker_memory_overhead)
    driver_res = WorkerResources(driver_cores or worker_cores, driver_memory or worker_memory, driver_num or worker_num,
                                 driver_cores_overhead or worker_cores_overhead, driver_timeout or worker_timeout,
                                 worker_memory_overhead)
    dedicated_operation_mode = dedicated_operation_mode and driver_num > 0

    enablers = enablers or SpytEnablers()

    spark_discovery = SparkDiscovery(discovery_path=discovery_path)

    current_operation_id = SparkDiscovery.getOption(spark_discovery.operation(), client=client)
    if current_operation_id is not None and get_operation_state(current_operation_id, client=client).is_running():
        if abort_existing:
            abort_spark_operations(spark_discovery, client)
        else:
            raise RuntimeError("This spark cluster is started already, use --abort-existing for auto restarting")

    ytserver_proxy_path = latest_ytserver_proxy_path(spark_cluster_version, client=client)
    global_conf = read_global_conf(client=client)
    if spark_cluster_version is None:
        spark_cluster_version = __scala_version__
    logger.info(f"{spark_cluster_version} cluster version will be launched")

    if enable_history_server and not advanced_event_log:
        logger.info("Reading event logs from dynamic tables is disabled in history server. "
                    "Applications use dynamic tables for logs by default, so history server may not work properly")

    if ssd_limit is not None:
        worker_disk_name = "ssd_slots_physical"
        worker_disk_limit = ssd_limit
        worker_disk_account = ssd_account

    if enable_tmpfs:
        logger.info("Tmpfs is enabled, spills will be created at RAM")

    if worker_disk_limit is None:
        logger.info("No disk account is specified")
        if enable_tmpfs:
            logger.info("Launcher files will be placed to tmpfs")
        else:
            logger.info("Launcher files will be placed to node disk with no guarantees on free space")

    validate_cluster_version(spark_cluster_version, client=client)
    validate_custom_params(params)
    validate_mtn_config(enablers, network_project, tvm_id, tvm_secret)
    validate_worker_num(worker_res.num, worker_num_limit(global_conf))
    validate_ssd_config(worker_disk_limit, worker_disk_account)

    dynamic_config = get_base_cluster_config(global_conf, spark_cluster_version, params,
                                             spark_discovery.base_discovery_path, client)

    _process_ipv6_preference(dynamic_config, enable_preference_ipv6, enablers)

    if ytserver_proxy_path:
        dynamic_config["ytserver_proxy_path"] = ytserver_proxy_path
    dynamic_config['spark_conf']['spark.dedicated_operation_mode'] = dedicated_operation_mode
    if enable_ytsaurus_shuffle:
        dynamic_config['spark_conf']['spark.shuffle.manager'] = \
            'org.apache.spark.shuffle.ytsaurus.YTsaurusShuffleManager'
        dynamic_config['spark_conf']['spark.shuffle.sort.io.plugin.class'] = \
            'tech.ytsaurus.spyt.shuffle.YTsaurusShuffleDataIO'
    else:
        dynamic_config['spark_conf']['spark.shuffle.service.enabled'] = 'true'
        set_random_shuffle_service_port(dynamic_config['spark_conf'])
    if autoscaler_period:
        dynamic_config['spark_conf']['spark.autoscaler.enabled'] = True
        dynamic_config['spark_conf']['spark.autoscaler.period'] = autoscaler_period
        if autoscaler_sliding_window:
            dynamic_config['spark_conf']['spark.autoscaler.sliding_window_size'] = autoscaler_sliding_window
        if autoscaler_max_free_workers:
            dynamic_config['spark_conf']['spark.autoscaler.max_free_workers'] = autoscaler_max_free_workers
        if autoscaler_slot_increment_step:
            dynamic_config['spark_conf']['spark.autoscaler.slots_increment_step'] = autoscaler_slot_increment_step

    enablers.apply_config(dynamic_config)

    spark_discovery.create(client)

    job_types = ['master']
    if enable_history_server:
        job_types.append('history')
    if enable_livy:
        job_types.append('livy')
    if not enable_multi_operation_mode:
        job_types.append('worker')

    common_config = CommonComponentConfig(
        operation_alias, pool, enable_tmpfs, network_project, tvm_id, tvm_secret, enablers, preemption_mode,
        cluster_log_level, rpc_job_proxy, rpc_job_proxy_thread_pool_size, enable_ytsaurus_shuffle,
        tcp_proxy_range_start, tcp_proxy_range_size, enable_stderr_table, "spark", spark_discovery, group_id,
        cluster_java_home
    )
    master_config = MasterConfig(master_memory_limit, master_port)
    worker_config = WorkerConfig(
        tmpfs_limit, worker_res, worker_port, None, True, autoscaler_period is not None, worker_log_transfer,
        worker_log_json_mode, worker_log_update_interval, worker_log_table_ttl, worker_disk_name,
        worker_disk_limit, worker_disk_account, worker_gpu_limit, cuda_toolkit_version(global_conf)
    )
    hs_config = HistoryServerConfig(
        history_server_memory_limit, history_server_cpu_limit, history_server_memory_overhead, shs_location,
        advanced_event_log
    )
    livy_config = LivyConfig(livy_driver_cores, livy_driver_memory, livy_max_sessions,
                             network_project=network_project)
    args = {
        'config': dynamic_config,
        'client': client,
        'job_types': job_types,
        'common_config': common_config,
        'master_config': master_config,
        'worker_config': worker_config,
        'hs_config': hs_config,
        'livy_config': livy_config,
    }

    master_args = args.copy()
    master_builder = build_spark_operation_spec(**master_args)

    op_child = None
    op_driver = None
    op = None

    try:
        op = run_operation(master_builder, sync=False, client=client)
        _wait_master_start(op, spark_discovery, client)
        logger.info("Master operation %s", op.id)

        if enable_multi_operation_mode:
            child_args = args.copy()
            child_args['job_types'] = ['worker']
            child_builder = build_spark_operation_spec(**child_args)
            op_child = run_operation(child_builder, sync=False, client=client)
            _wait_child_start(op_child, spark_discovery, client)
            logger.info("Child operation %s", op_child.id)

        if dedicated_operation_mode:
            driver_args = args.copy()
            driver_args['job_types'] = ['driver']
            driver_config = worker_config._replace(res=driver_res,
                                                   driver_op_discovery_script='bin/driver-op-discovery.sh')
            driver_args['worker_config'] = driver_config
            driver_builder = build_spark_operation_spec(**driver_args)
            op_driver = run_operation(driver_builder, sync=False, client=client)
            _wait_child_start(op_driver, spark_discovery, client)
            logger.info("Driver operation %s", op_driver.id)

        logger.info("Spark Master's Web UI: {0}".format(find_spark_cluster(discovery_path, client).master_webui()))
        return op
    except Exception as err:
        logging.error(err, exc_info=True)
        abort_operation_silently(op_driver, client)
        abort_operation_silently(op_child, client)
        abort_operation_silently(op, client)


def abort_operation_silently(op_id, client):
    try:
        if op_id:
            abort_operation(op_id, client=client)
    except Exception as err:
        logging.error("Failed to abort operation {0}".format(op_id), exc_info=True)
        return err


def find_spark_cluster(discovery_path=None, client=None):
    """Print Spark urls
    :param discovery_path: Cypress path for discovery files and logs
    :param client: YtClient
    :return: None
    """
    discovery = SparkDiscovery(discovery_path=discovery_path)
    master_jobs_path = discovery.discovery().join("master_jobs")
    master_jobs = SparkDiscovery.getOptions(master_jobs_path, client=client)
    master_job = yt_get(master_jobs_path.join(master_jobs[0]), client=client) if master_jobs else None
    return SparkCluster(
        master_endpoint=SparkDiscovery.getOption(discovery.master_spark(), client=client),
        master_web_ui_url=SparkDiscovery.getOption(discovery.master_webui(), client=client),
        master_web_ui_schemed_url=master_job["webui_url"] if master_job else None,
        master_rest_endpoint=SparkDiscovery.getOption(discovery.master_rest(), client=client),
        operation_id=SparkDiscovery.getOption(discovery.operation(), client=client),
        shs_url=SparkDiscovery.getOption(discovery.shs(), client=client),
        livy_url=SparkDiscovery.getOption(discovery.livy(), client=client),
        spark_cluster_version=SparkDiscovery.getOption(discovery.spark_cluster_version(), client=client),
        children_operation_ids=SparkDiscovery.getOptions(discovery.children_operations(), client=client)
    )



import copy
import logging
import os
import shlex
from dataclasses import dataclass
from typing import List, NamedTuple
from pyspark import __version__ as spark_version

from spyt.dependency_utils import require_yt_client

require_yt_client()

from yt.wrapper import YtClient  # noqa: E402
from yt.wrapper.common import update_inplace, update  # noqa: E402
from yt.wrapper.http_helpers import get_token, get_user_name  # noqa: E402
from yt.wrapper.spec_builders import VanillaSpecBuilder  # noqa: E402

from .conf import ytserver_proxy_attributes, get_spark_distributive  # noqa: E402
from .conf import read_metrics_conf, read_spark_defaults_conf  # noqa: E402
from .utils import SparkDiscovery, call_get_proxy_address_url, parse_memory  # noqa: E402
from .enabler import SpytEnablers  # noqa: E402
from .version import __version__  # noqa: E402

logger = logging.getLogger(__name__)


class SparkDefaultArguments(object):
    SPARK_WORKER_MEMORY_OVERHEAD = "2G"
    SPARK_WORKER_TMPFS_LIMIT = "8G"
    SPARK_WORKER_SSD_LIMIT = None
    SPARK_MASTER_PORT = 27001
    SPARK_WORKER_PORT = 27001
    SPARK_MASTER_MEMORY_LIMIT = "4G"
    SPARK_HISTORY_SERVER_MEMORY_LIMIT = "4G"
    SPARK_HISTORY_SERVER_MEMORY_OVERHEAD = "2G"
    SPARK_HISTORY_SERVER_CPU_LIMIT = 1
    SPARK_WORKER_TIMEOUT = "10m"
    SPARK_WORKER_LOG_UPDATE_INTERVAL = "10m"
    SPARK_WORKER_LOG_TABLE_TTL = "7d"
    SPARK_WORKER_CORES_OVERHEAD = 0
    SPARK_WORKER_CORES_BYOP_OVERHEAD = 0
    SPARK_WORKER_GPU_LIMIT = 0
    LIVY_DRIVER_CORES = 1
    LIVY_DRIVER_MEMORY = "1024m"
    LIVY_MAX_SESSIONS = 3

    @staticmethod
    def get_params():
        return {
            "operation_spec": {
                "annotations": {
                    "is_spark": True
                },
                "max_failed_job_count": 10000,
                "max_stderr_count": 150,
                "job_cpu_monitor": {
                    "enable_cpu_reclaim": False
                }
            }
        }


@dataclass
class CommonComponentConfig():
    operation_alias: str = None
    pool: str = None
    enable_tmpfs: bool = True
    network_project: str = None
    tvm_id: str = None
    tvm_secret: str = None
    enablers: SpytEnablers = None
    preemption_mode: str = "normal"
    cluster_log_level: str = "INFO"
    rpc_job_proxy: bool = False
    rpc_job_proxy_thread_pool_size: int = 4
    enable_ytsaurus_shuffle: bool = False
    tcp_proxy_range_start: int = 30000
    tcp_proxy_range_size: int = 100
    enable_stderr_table: bool = False
    alias_prefix: str = ""
    spark_discovery: SparkDiscovery = None
    group_id: str = None
    cluster_java_home: str = None
    container_home: str = None
    spark_home: str = None
    spyt_home: str = None

    def __post_init__(self):
        enable_squashfs = self.enablers and self.enablers.enable_squashfs
        if not self.container_home:
            self.container_home = "./tmpfs" if self.enable_tmpfs else "."
        self.spark_home = "/usr/lib/spark" if enable_squashfs else f"$HOME/{self.container_home}/spark"
        self.spyt_home = "/usr/lib/spyt" if enable_squashfs else f"$HOME/{self.container_home}/spyt-package"


class MasterConfig(NamedTuple):
    master_memory_limit: str = SparkDefaultArguments.SPARK_MASTER_MEMORY_LIMIT
    master_port: int = SparkDefaultArguments.SPARK_MASTER_PORT


class WorkerResources(NamedTuple):
    cores: int
    memory: str
    num: int
    cores_overhead: int
    timeout: str
    memory_overhead: str


class WorkerConfig(NamedTuple):
    tmpfs_limit: str = SparkDefaultArguments.SPARK_WORKER_TMPFS_LIMIT
    res: WorkerResources = None
    worker_port: int = SparkDefaultArguments.SPARK_WORKER_PORT
    driver_op_discovery_script: str = None
    extra_metrics_enabled: bool = True
    autoscaler_enabled: bool = False
    worker_log_transfer: bool = False
    worker_log_json_mode: bool = False
    worker_log_update_interval: str = SparkDefaultArguments.SPARK_WORKER_LOG_UPDATE_INTERVAL
    worker_log_table_ttl: str = SparkDefaultArguments.SPARK_WORKER_LOG_TABLE_TTL
    worker_disk_name: str = "default"
    worker_disk_limit: str = None
    worker_disk_account: str = None
    worker_gpu_limit: int = 0
    cuda_toolkit_version: str = None


class HistoryServerConfig(NamedTuple):
    history_server_memory_limit: str = SparkDefaultArguments.SPARK_HISTORY_SERVER_MEMORY_LIMIT
    history_server_cpu_limit: int = SparkDefaultArguments.SPARK_HISTORY_SERVER_CPU_LIMIT
    history_server_memory_overhead: str = SparkDefaultArguments.SPARK_HISTORY_SERVER_MEMORY_OVERHEAD
    shs_location: str = None
    advanced_event_log: bool = False


class LivyConfig(NamedTuple):
    livy_driver_cores: int = SparkDefaultArguments.LIVY_DRIVER_CORES
    livy_driver_memory: str = SparkDefaultArguments.LIVY_DRIVER_MEMORY
    livy_max_sessions: int = SparkDefaultArguments.LIVY_MAX_SESSIONS
    spark_master_address: str = None
    master_group_id: str = None
    network_project: str = None


class CommonSpecParams(NamedTuple):
    spark_distributive: str
    java_home: str
    extra_java_opts: List[str]
    environment: dict
    spark_conf: dict
    task_spec: dict
    config: CommonComponentConfig
    entrypoint: str = ''


def _put_if_not_none(d, key, value):
    if value is not None:
        d[key] = value


def spark_conf_to_opts(config: dict):
    return ["-D{}={}".format(key, value) for key, value in config.items()]


def setup_spyt_env(container_home: str, spark_distributive: str, enable_squashfs: bool,
                   additional_parameters: List[str]):
    cmd = ["./setup-spyt-env.sh"]
    if enable_squashfs:
        cmd += ["--use-squashfs"]
    else:
        cmd += ["--spark-home", container_home, "--spark-distributive", spark_distributive]
    return " ".join(cmd + additional_parameters)


def _launcher_command(component: str, common_params: CommonSpecParams, additional_parameters: List[str] = None,
                      xmx: str = "512m", extra_java_opts: List[str] = None, launcher_opts: str = ""):
    additional_parameters = additional_parameters or []
    extra_java_opts = extra_java_opts or []
    setup_spyt_env_cmd = setup_spyt_env(common_params.config.container_home,
                                        common_params.spark_distributive,
                                        common_params.config.enablers.enable_squashfs,
                                        additional_parameters)

    java_bin = os.path.join(common_params.java_home, 'bin', 'java') if common_params.java_home else 'java'
    if command := common_params.entrypoint:
        java_bin = f'{command} {java_bin}'
    classpath = (f'{common_params.config.spyt_home}/conf/:'
                 f'{common_params.config.spyt_home}/jars/*:'
                 f'{common_params.config.spark_home}/jars/*')
    extra_java_opts_str = " ".join(extra_java_opts)
    run_launcher = "{} -Xmx{} -cp {} {}".format(java_bin, xmx, classpath, extra_java_opts_str)

    commands = [
        setup_spyt_env_cmd,
        "{} tech.ytsaurus.spark.launcher.{}Launcher {}".format(run_launcher, component, launcher_opts)
    ]
    return " && ".join(commands)


def _create_file_and_layer_paths(config, enable_squashfs: bool, spark_distributive_path):
    file_paths = [path for path in config["file_paths"] if
                  not enable_squashfs or
                  not path.endswith('spyt-package.zip')]
    layer_paths = []

    if enable_squashfs:
        layer_paths.append(f"{config['spark_yt_base_path']}/spyt-package.squashfs")
        layer_paths.append(spark_distributive_path)
    else:
        file_paths.append(spark_distributive_path)

    layer_paths += config["squashfs_layer_paths" if enable_squashfs else "layer_paths"]
    return file_paths, layer_paths


def build_livy_spec(builder: VanillaSpecBuilder, common_params: CommonSpecParams, config: LivyConfig):
    livy_launcher_opts = [
        f"--driver-cores {config.livy_driver_cores}",
        f"--driver-memory {config.livy_driver_memory}",
        f"--max-sessions {config.livy_max_sessions}",
    ]
    enable_squashfs = common_params.config.enablers.enable_squashfs
    if config.spark_master_address is not None:
        livy_launcher_opts.append(f"--master-address {config.spark_master_address}")
    if config.network_project is not None:
        livy_launcher_opts.append(f"--network-project {config.network_project}")
    if enable_squashfs:
        livy_launcher_opts.append("--enable-squashfs")
    extra_java_opts = common_params.extra_java_opts + spark_conf_to_opts(common_params.spark_conf)
    livy_command = _launcher_command("Livy", common_params, additional_parameters=["--enable-livy"],
                                     extra_java_opts=extra_java_opts, launcher_opts=" ".join(livy_launcher_opts))

    livy_work_dir = "$HOME/{}/livy".format(common_params.config.container_home)
    livy_environment = {
        "LIVY_HOME": "/usr/lib/livy" if enable_squashfs else livy_work_dir,
        "LIVY_WORK_DIR": livy_work_dir
    }
    _put_if_not_none(livy_environment, "SPARK_MASTER_DISCOVERY_GROUP_ID", config.master_group_id)
    livy_environment = update(common_params.environment, livy_environment)

    livy_layer_paths = []
    livy_file_paths = copy.copy(common_params.task_spec["file_paths"])

    if enable_squashfs:
        livy_layer_paths.append("//home/spark/livy/livy.squashfs")
    else:
        livy_file_paths.append("//home/spark/livy/livy.tgz")

    livy_layer_paths += common_params.task_spec["layer_paths"]

    livy_task_spec = copy.deepcopy(common_params.task_spec)
    livy_task_spec["environment"] = livy_environment
    livy_task_spec["file_paths"] = livy_file_paths
    livy_task_spec["layer_paths"] = livy_layer_paths

    builder.begin_task("livy") \
        .job_count(1) \
        .command(livy_command) \
        .memory_limit(parse_memory("1G") +
                      parse_memory(config.livy_driver_memory) * config.livy_max_sessions) \
        .cpu_limit(1 + config.livy_driver_cores * config.livy_max_sessions) \
        .spec(livy_task_spec) \
        .end_task()


def build_hs_spec(builder: VanillaSpecBuilder, common_params: CommonSpecParams, config: HistoryServerConfig):
    spark_conf_hs = common_params.spark_conf.copy()
    if "spark.workerLog.tablePath" not in spark_conf_hs:
        worker_log_location = "yt:/{}".format(common_params.config.spark_discovery.worker_log())
        spark_conf_hs["spark.workerLog.tablePath"] = worker_log_location
    if config.advanced_event_log:
        event_log_path = "ytEventLog:/{}".format(config.shs_location or
                                                 common_params.config.spark_discovery.event_log_table())
    else:
        event_log_path = "yt:/{}".format(config.shs_location or common_params.config.spark_discovery.event_log())
    if "spark.history.fs.numReplayThreads" not in spark_conf_hs:
        spark_conf_hs["spark.history.fs.numReplayThreads"] = config.history_server_cpu_limit

    history_launcher_opts = "--log-path {} --memory {}".format(event_log_path, config.history_server_memory_limit)
    extra_java_opts = common_params.extra_java_opts + spark_conf_to_opts(spark_conf_hs)
    history_command = _launcher_command("HistoryServer", common_params, extra_java_opts=extra_java_opts,
                                        launcher_opts=history_launcher_opts)

    shs_file_paths = copy.copy(common_params.task_spec["file_paths"])

    builder.begin_task("history") \
        .job_count(1) \
        .command(history_command) \
        .memory_limit(parse_memory(config.history_server_memory_limit) +
                      parse_memory(config.history_server_memory_overhead)) \
        .cpu_limit(config.history_server_cpu_limit) \
        .spec(common_params.task_spec) \
        .file_paths(shs_file_paths) \
        .end_task()


def build_master_spec(builder: VanillaSpecBuilder, common_params: CommonSpecParams, config: MasterConfig):
    extra_java_opts = common_params.extra_java_opts + spark_conf_to_opts(common_params.spark_conf)
    master_command = _launcher_command("Master", common_params, extra_java_opts=extra_java_opts)
    master_environment = {
        "SPARK_MASTER_PORT": str(config.master_port),
    }
    master_environment = update(common_params.environment, master_environment)
    master_task_spec = copy.deepcopy(common_params.task_spec)
    master_task_spec["environment"] = master_environment
    task = builder.begin_task("master") \
        .job_count(1) \
        .command(master_command) \
        .memory_limit(parse_memory(config.master_memory_limit)) \
        .cpu_limit(2) \
        .spec(master_task_spec)
    if common_params.spark_conf.get("spark.ytsaurus.use_assigned_ports.master") == "true":
        task.port_count(2)
    task.end_task()


def build_worker_spec(builder: VanillaSpecBuilder, job_type: str, ytserver_proxy_path: str, tvm_enabled: bool,
                      common_params: CommonSpecParams, config: WorkerConfig):
    def _script_absolute_path(script):
        return os.path.join(common_params.config.spyt_home, script)

    spark_conf_worker = common_params.spark_conf.copy()

    if "spark.workerLog.tablePath" not in spark_conf_worker:
        worker_log_location = "yt:/{}".format(common_params.config.spark_discovery.worker_log())
        spark_conf_worker["spark.workerLog.tablePath"] = worker_log_location
    if config.driver_op_discovery_script:
        driver_op_discovery_script_path = _script_absolute_path(config.driver_op_discovery_script)
        spark_conf_worker["spark.worker.resource.driverop.amount"] = str(config.res.cores)
        spark_conf_worker["spark.worker.resource.driverop.discoveryScript"] = driver_op_discovery_script_path
        spark_conf_worker["spark.driver.resource.driverop.discoveryScript"] = driver_op_discovery_script_path
    if config.extra_metrics_enabled:
        spark_conf_worker["spark.ui.prometheus.enabled"] = "true"

    if config.autoscaler_enabled:
        job_id_script_path = _script_absolute_path("bin/job-id-discovery.sh")
        spark_conf_worker["spark.worker.resource.jobid.discoveryScript"] = job_id_script_path
        spark_conf_worker["spark.driver.resource.jobid.discoveryScript"] = job_id_script_path
        spark_conf_worker["spark.worker.resource.jobid.amount"] = "1"

    if config.worker_gpu_limit > 0:
        spark_conf_worker["spark.worker.resource.gpu.amount"] = str(config.worker_gpu_limit)
        spark_conf_worker["spark.worker.resource.gpu.discoveryScript"] = _script_absolute_path("bin/getGpusResources.sh")

    worker_launcher_opts = \
        "--cores {0} --memory {1} --wait-master-timeout {2} --wlog-service-enabled {3} --wlog-enable-json {4} " \
        "--wlog-update-interval {5} --wlog-table-ttl {6} {7}".format(
            config.res.cores, config.res.memory,
            config.res.timeout, config.worker_log_transfer, config.worker_log_json_mode,
            config.worker_log_update_interval, config.worker_log_table_ttl,
            "--enable-squashfs" if common_params.config.enablers.enable_squashfs else ""
        )
    extra_java_opts = common_params.extra_java_opts + spark_conf_to_opts(spark_conf_worker)
    worker_command = _launcher_command("Worker", common_params, xmx="2g", extra_java_opts=extra_java_opts,
                                       launcher_opts=worker_launcher_opts)

    worker_environment = {
        "SPARK_YT_BYOP_ENABLED": str(common_params.config.enablers.enable_byop),
        "SPARK_WORKER_PORT": str(config.worker_port),
        "SPARK_YT_CLUSTER_CONF_PATH": str(common_params.config.spark_discovery.conf()),
    }
    worker_environment = update(common_params.environment, worker_environment)
    if common_params.config.enablers.enable_byop:
        ytserver_binary_name = ytserver_proxy_path.split("/")[-1] if ytserver_proxy_path else "ytserver-proxy"
        byop_worker_environment = {
            "SPARK_YT_BYOP_BINARY_PATH": "$HOME/{}".format(ytserver_binary_name),
            "SPARK_YT_BYOP_CONFIG_PATH": "$HOME/ytserver-proxy.template.yson",
            "SPARK_YT_BYOP_HOST": "localhost",
            "SPARK_YT_BYOP_TVM_ENABLED": str(tvm_enabled)
        }
        worker_environment = update(worker_environment, byop_worker_environment)

    if common_params.config.enablers.enable_byop:
        worker_cores_overhead = config.res.cores_overhead or SparkDefaultArguments.SPARK_WORKER_CORES_BYOP_OVERHEAD
    else:
        worker_cores_overhead = config.res.cores_overhead or SparkDefaultArguments.SPARK_WORKER_CORES_OVERHEAD

    worker_file_paths = copy.copy(common_params.task_spec["file_paths"])
    if ytserver_proxy_path and common_params.config.enablers.enable_byop:
        worker_file_paths.append(ytserver_proxy_path)

    worker_task_spec = copy.deepcopy(common_params.task_spec)
    worker_task_spec["environment"] = worker_environment
    worker_task_spec["rpc_proxy_worker_thread_pool_size"] = common_params.config.rpc_job_proxy_thread_pool_size
    worker_task_spec["enable_shuffle_service_in_job_proxy"] = common_params.config.enable_ytsaurus_shuffle
    worker_ram_memory = parse_memory(config.res.memory)
    worker_local_dirs = "."
    if config.worker_disk_limit:
        worker_task_spec["disk_request"] = {
            "disk_space": parse_memory(config.worker_disk_limit),
            "account": config.worker_disk_account,
            "medium_name": config.worker_disk_name
        }
    elif common_params.config.enable_tmpfs:
        tmpfs_limit = parse_memory(config.tmpfs_limit)
        worker_ram_memory += tmpfs_limit
        worker_task_spec['tmpfs_size'] = tmpfs_limit
        worker_local_dirs = "./tmpfs"
    worker_environment["SPARK_LOCAL_DIRS"] = worker_local_dirs

    if config.worker_gpu_limit > 0:
        worker_task_spec["cuda_toolkit_version"] = config.cuda_toolkit_version
        worker_task_spec["gpu_limit"] = config.worker_gpu_limit

    driver_task_spec = copy.deepcopy(worker_task_spec)
    driver_environment = driver_task_spec["environment"]
    if config.driver_op_discovery_script:
        driver_environment["SPARK_DRIVER_RESOURCE"] = str(config.res.cores)

    spec = driver_task_spec if "drivers" == job_type else worker_task_spec
    task = builder.begin_task(job_type) \
        .job_count(config.res.num) \
        .command(worker_command) \
        .memory_limit(worker_ram_memory + parse_memory(config.res.memory_overhead)) \
        .cpu_limit(config.res.cores + worker_cores_overhead) \
        .spec(spec) \
        .file_paths(worker_file_paths)
    if common_params.spark_conf.get("spark.ytsaurus.use_assigned_ports.worker") == "true":
        task.port_count(2)
    task.end_task()


def build_spark_operation_spec(config: dict, client: YtClient,
                               job_types: List[str], common_config: CommonComponentConfig,
                               master_config: MasterConfig = None, worker_config: WorkerConfig = None,
                               hs_config: HistoryServerConfig = None, livy_config: LivyConfig = None):
    if job_types == [] or job_types is None:
        job_types = ['master', 'history', 'worker']

    spark_distributive, spark_distributive_path = get_spark_distributive(client, common_config.enablers.enable_squashfs)

    extra_java_opts = ["-Dlog4j.loglevel={}".format(common_config.cluster_log_level)]
    if common_config.enablers.enable_preference_ipv6:
        extra_java_opts.append("-Djava.net.preferIPv6Addresses=true")

    spark_conf_common = config["spark_conf"].copy()
    update_inplace(spark_conf_common, common_config.enablers.get_conf())

    user = get_user_name(client=client)

    operation_spec = copy.copy(config["operation_spec"])
    if common_config.spark_discovery is not None and common_config.enable_stderr_table:
        if "master" in job_types:
            operation_spec["stderr_table_path"] = str(common_config.spark_discovery.stderr())
        elif "driver" in job_types:
            operation_spec["stderr_table_path"] = str(common_config.spark_discovery.stderr() + "_driver")
        else:
            operation_spec["stderr_table_path"] = str(common_config.spark_discovery.stderr()) + "_worker"
    if "title" not in operation_spec:
        operation_spec["title"] = common_config.operation_alias or (common_config.alias_prefix + "_" + str(user))
    operation_spec["pool"] = common_config.pool
    operation_spec['preemption_mode'] = common_config.preemption_mode
    description = {
        "cluster_version": config["cluster_version"],
        "client_version": __version__,
        "enablers": str(common_config.enablers),
        "job_types": job_types,
    }
    if common_config.spark_discovery is not None:
        description["discovery_path"] = common_config.spark_discovery.base_discovery_path
    _put_if_not_none(description, "discovery_group_id", common_config.group_id)
    operation_spec["description"] = {
        "Spark over YT": description
    }
    ytserver_proxy_path = config.get("ytserver_proxy_path")
    if ytserver_proxy_path and common_config.enablers.enable_byop:
        operation_spec["description"]["BYOP"] = ytserver_proxy_attributes(ytserver_proxy_path, client=client)

    environment = copy.deepcopy(config["environment"])
    java_home = common_config.cluster_java_home or config.get('default_cluster_java_home')
    if "JAVA_HOME" in environment:
        del environment["JAVA_HOME"]  # COMPAT(atokarew) JAVA_HOME is preserved in global config for older releases
    if java_home:
        environment["JAVA_HOME"] = java_home
    environment["YT_PROXY"] = call_get_proxy_address_url(required=True, client=client)
    environment["YT_OPERATION_ALIAS"] = operation_spec["title"]
    _put_if_not_none(environment, "SPARK_DISCOVERY_GROUP_ID", common_config.group_id)
    if common_config.spark_discovery is not None:
        environment["SPARK_BASE_DISCOVERY_PATH"] = str(common_config.spark_discovery.base_discovery_path)
        environment["SPARK_DISCOVERY_PATH"] = str(common_config.spark_discovery.discovery())  # COMPAT(alex-shishkin)
    environment["SPARK_HOME"] = common_config.spark_home
    environment["SPYT_HOME"] = common_config.spyt_home
    environment["SPARK_CLUSTER_VERSION"] = config["cluster_version"]  # TODO Rename to SPYT_CLUSTER_VERSION
    environment["SPYT_CLUSTER_VERSION"] = config["cluster_version"]
    environment["SPARK_YT_METRICS_ENABLED"] = str(common_config.enablers.enable_yt_metrics)
    environment["SPARK_YT_IPV6_PREFERENCE_ENABLED"] = \
        str(common_config.enablers.enable_preference_ipv6)  # COMPAT(alex-shishkin)
    environment["SPARK_YT_TCP_PROXY_ENABLED"] = str(common_config.enablers.enable_tcp_proxy)
    environment["SPARK_YT_PROFILING_ENABLED"] = str(common_config.enablers.enable_profiling)
    environment["SPARK_YT_RPC_JOB_PROXY_ENABLED"] = str(common_config.rpc_job_proxy)
    if common_config.enablers.enable_byop:
        environment["SPARK_YT_BYOP_PORT"] = "27002"
    if common_config.enablers.enable_yt_metrics:
        metrics_conf = read_metrics_conf()
        spark_defaults_conf = read_spark_defaults_conf()

        spark_push_port = metrics_conf["*.sink.solomon.solomon_port"]
        metrics_pull_port = spark_defaults_conf["spark.ytsaurus.metrics.pull.port"]
        agent_pull_port = spark_defaults_conf["spark.ytsaurus.metrics.agent.pull.port"]
        environment["YT_METRICS_SPARK_PUSH_PORT"] = spark_push_port
        environment["YT_AGENT_METRICS_PULL_PORT"] = agent_pull_port
        environment["YT_METRICS_PULL_PORT"] = metrics_pull_port
        operation_spec["annotations"]["solomon_resolver_tag"] = "spark"
        operation_spec["annotations"]["solomon_resolver_ports"] = [int(metrics_pull_port), int(agent_pull_port)]

    if common_config.enablers.enable_tcp_proxy:
        environment["SPARK_YT_TCP_PROXY_RANGE_START"] = str(common_config.tcp_proxy_range_start)
        environment["SPARK_YT_TCP_PROXY_RANGE_SIZE"] = str(common_config.tcp_proxy_range_size)

    file_paths, layer_paths = _create_file_and_layer_paths(config,
                                                           common_config.enablers.enable_squashfs,
                                                           spark_distributive_path)

    common_task_spec = {
        "restart_completed_jobs": True,
        "file_paths": file_paths,
        "layer_paths": layer_paths,
        "environment": environment,
        "memory_reserve_factor": 1.0,
        "enable_rpc_proxy_in_job_proxy": common_config.rpc_job_proxy,
    }
    if common_config.enable_tmpfs:
        common_task_spec["tmpfs_path"] = "tmpfs"
    if common_config.enablers.enable_mtn:
        common_task_spec["network_project"] = common_config.network_project
    if common_config.group_id:
        common_task_spec['extra_environment'] = ['discovery_server_addresses']

    tvm_enabled = common_config.enablers.enable_mtn and bool(common_config.tvm_id) and bool(common_config.tvm_secret)
    secure_vault = {"YT_USER": user, "YT_TOKEN": get_token(client=client)}
    if tvm_enabled:
        secure_vault["SPARK_TVM_ID"] = common_config.tvm_id
        secure_vault["SPARK_TVM_SECRET"] = common_config.tvm_secret

    entrypoint = config.get('entrypoint', '')
    if not isinstance(entrypoint, str):
        entrypoint = shlex.join(entrypoint)
    common_params = CommonSpecParams(
        spark_distributive, java_home, extra_java_opts, environment,
        spark_conf_common, common_task_spec, common_config, entrypoint,
    )
    builder = VanillaSpecBuilder()
    if "master" in job_types:
        build_master_spec(builder, common_params, master_config)
    if "history" in job_types:
        build_hs_spec(builder, common_params, hs_config)
    if "worker" in job_types:
        build_worker_spec(builder, "workers", ytserver_proxy_path, tvm_enabled, common_params, worker_config)
    if "driver" in job_types:
        build_worker_spec(builder, "drivers", ytserver_proxy_path, tvm_enabled, common_params, worker_config)
    if "livy" in job_types:
        build_livy_spec(builder, common_params, livy_config)

    return builder \
        .secure_vault(secure_vault) \
        .spec(operation_spec)


def build_spark_connect_server_spec(client: YtClient, config, spark_conf: dict, enablers: SpytEnablers, java_home: str,
                                    driver_memory: str, num_executors: int, executor_cores: int, executor_memory: str,
                                    prefer_ipv6: bool, pool: str, grpc_port_start: int):
    component_config = CommonComponentConfig(enable_tmpfs=False, rpc_job_proxy=True, enablers=enablers)

    spark_distributive, spark_distributive_path = get_spark_distributive(client, enablers.enable_squashfs)
    setup = setup_spyt_env(component_config.container_home, spark_distributive, enablers.enable_squashfs, [])
    user = get_user_name(client=client)
    yt_proxy = call_get_proxy_address_url(required=True, client=client)
    spark_connect_jar = f"spark-connect_2.12-{spark_version}.jar"
    network_project = spark_conf.get("spark.ytsaurus.network.project")
    command = [
        f"{component_config.spark_home}/bin/spark-submit",
        f"--master ytsaurus://{yt_proxy}",
        "--deploy-mode client",
        "--class org.apache.spark.sql.connect.ytsaurus.SpytConnectServer",
        f"--num-executors {num_executors}",
        f"--executor-cores {executor_cores}",
        f"--executor-memory {executor_memory}",
        f"--queue {pool or user}",
        f'--name "Spark connect driver for {user}"',
        f"--conf spark.driver.extraJavaOptions='-Djava.net.preferIPv6Addresses={prefer_ipv6}'",
        f"--conf spark.connect.grpc.binding.port={grpc_port_start}",
        f"--jars {spark_connect_jar}",
    ] + [f"--conf {key}={spark_conf[key]}" for key in spark_conf] + ["spark-internal"]

    operation_spec = {
        "title": "Spark connect server"
    }

    secure_vault = {
        "YT_TOKEN": get_token(client=client)
    }

    file_paths, layer_paths = _create_file_and_layer_paths(config, enablers.enable_squashfs, spark_distributive_path)

    file_paths.append(f"//home/spark/distrib/{spark_version.replace('.', '/')}/spark-connect_2.12-{spark_version}.jar")

    environment = copy.deepcopy(config["environment"])
    environment["JAVA_HOME"] = java_home
    environment["SPARK_CONF_DIR"] = f"{component_config.spyt_home}/conf".replace("$HOME/", "")
    environment["SPARK_CONNECT_CLASSPATH"] = f"{spark_connect_jar}"

    task_spec = {
        "layer_paths": layer_paths,
        "file_paths": file_paths,
        "enable_rpc_proxy_in_job_proxy": component_config.rpc_job_proxy,
        "environment": environment,
    }

    if network_project:
        task_spec["network_project"] = network_project

    builder = VanillaSpecBuilder()
    builder.begin_task("spark_connect_driver") \
        .command(" ".join([setup, "&&"] + command)) \
        .job_count(1) \
        .cpu_limit(1) \
        .memory_limit(parse_memory(driver_memory)) \
        .spec(task_spec) \
        .end_task()

    return builder.secure_vault(secure_vault) \
        .spec(operation_spec)



"""SPYT extensions for pyspark module.

Usage notes: spyt module must be imported before any of pyspark.* modules in order
for extensions to take an effect.
"""
import logging

from .dependency_utils import require_pyspark

require_pyspark()

from .client import connect, spark_session, connect_direct, direct_spark_session, \
    info, stop, jvm_process_pid, yt_client, is_stopped  # noqa: E402
from .extensions import read_yt, read_schema_hint, write_yt, sorted_by, optimize_for, withYsonColumn, transform, \
    write_schema_hint  # noqa: E402
from .utils import check_spark_version  # noqa: E402
from spyt.types import UInt64Type  # noqa: E402
import pyspark.context  # noqa: E402
import pyspark.sql.types  # noqa: E402
import pyspark.sql.readwriter  # noqa: E402
import pyspark.cloudpickle.cloudpickle  # noqa: E402
import pyspark.cloudpickle.cloudpickle_fast  # noqa: E402
from types import CodeType  # noqa: E402

__all__ = [
    'connect',
    'connect_direct',
    'spark_session',
    'direct_spark_session',
    'info',
    'stop',
    'jvm_process_pid',
    'yt_client',
    'is_stopped'
]


def configure_logging():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    logger.propagate = False

    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)

    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
    handler.setFormatter(formatter)

    logger.addHandler(handler)


def initialize():
    pyspark.sql.types._atomic_types.append(UInt64Type)
    # exact copy of the corresponding line in pyspark/sql/types.py
    pyspark.sql.types._all_atomic_types = dict((t.typeName(), t) for t in pyspark.sql.types._atomic_types)
    pyspark.sql.types._acceptable_types.update({UInt64Type: (int,)})

    pyspark.sql.readwriter.DataFrameReader.yt = read_yt
    pyspark.sql.readwriter.DataFrameReader.schema_hint = read_schema_hint

    pyspark.sql.readwriter.DataFrameWriter.yt = write_yt
    pyspark.sql.readwriter.DataFrameWriter.schema_hint = write_schema_hint
    pyspark.sql.readwriter.DataFrameWriter.sorted_by = sorted_by
    pyspark.sql.readwriter.DataFrameWriter.optimize_for = optimize_for

    pyspark.sql.dataframe.DataFrame.withYsonColumn = withYsonColumn
    pyspark.sql.dataframe.DataFrame.transform = transform

    pyspark.context.SparkContext.PACKAGE_EXTENSIONS += ('.whl',)

    if check_spark_version(less_than="3.4.0"):
        from .extensions import _extract_code_globals, _code_reduce

        pyspark.cloudpickle.cloudpickle._extract_code_globals = _extract_code_globals
        pyspark.cloudpickle.cloudpickle_fast._extract_code_globals = _extract_code_globals
        pyspark.cloudpickle.cloudpickle_fast._code_reduce = _code_reduce
        pyspark.cloudpickle.cloudpickle_fast.CloudPickler.dispatch_table[CodeType] = _code_reduce


configure_logging()
initialize()
